

/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef ATOM_H
#define ATOM_H

#include "StateVar.h"
#include "../system/purge.h"
#include <vector>

#if (__GNUC__ >= 3)
#include <ext/hash_map>
#else
#include <hash_map>
#endif

using namespace std;
#if (__GNUC__ >= 3)
using __gnu_cxx::hash_map;
using __gnu_cxx::hash;
#else
using std::hash_map;
using std::hash;
#endif

#ifndef EQSTR_S
#define EQSTR_S
struct eqstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) == 0;
  }
};
#endif

class Transition;
class Formula;
class Comp;

/**
* \class Atom
* \brief An atom is a litteral, i.e. a variable assignment, with associated truth values.
*/

class Atom {
  
 public:
  StateVar* stateVar;
  std::string value;
  std::string id;  /**< id of the atom = string of its propositional value (ex : vdu.mode = on)*/
  unsigned int dispid; /**< integer for display id */
  bool fail; /**< if the atom corresponds to a failure or not */
  std::vector<Transition*>* ref_trans_pre; /**< references to the transitions in which the atom belongs to the premisses */
  std::vector<Transition*>* ref_trans_post; /**< references to the transitions in which the atom belongs to the post effects */
  Clause* support; //itms labels and support on atoms.
  std::vector<Atom*>* supported; //other atoms currently supported by this one. Theroretically the current atom must be present in the support's clause of each of these atoms...
  std::vector<Clause*>* ref_clauses;
  hash_map<const char*, Comp*, hash<const char*>, eqstr>* ref_comps;
  std::string label; //label for propagation and conflicts resolution (true, false, unknown).
  int propagation_number; //propagation number for fast dependencies evaluation.
  bool neg; //negation (true means negative, false means positive).
  double post_prob; /**< probability from transitions to this atom. */
  bool cond; /**< the atom is linked to a condition. */
  Transition* idle;
  unsigned int guards_number;
  hash_map<const char*, std::string, hash<const char*>, eqstr>* dyn_name;
  bool dead_end;  /* no nominal transitions out of that mode */
  int topo_number; /* topological number */

 public:
  Atom(StateVar* staVar, const std::string &val, const std::string &ide);
  Atom(StateVar* staVar, const std::string &val, const std::string &ide, 
       const bool &non); //with negation choice.
  ~Atom();
  bool evalAtom();
  void set(); //set the atom true (modify the value of the corresponding state var).
  void setLabel(const std::string &lab); //set the label.
  void setLabel(const std::string &lab, const bool &update); //set the label and update or not the label of the others atoms of the state variable.
  void addRefConstPre(Formula* form); //add a formula to the ref of constraints.
  void addRefTransPre(Transition* trans); //add a transition to the ref of transitions.
  void addRefConstPost(Formula* form);
  void addRefTransPost(Transition* trans);
  void showAtom(bool ev); //bool ev, to show evaluation, or not.
  void decTopoNumber();
  void incTopoNumber();
  int getTopoNumber();

/* inline methods */
/**
 * @return the state variable associated to the atom.
 */
inline StateVar* Atom::getStateVar() {
  return stateVar;
}
/**
 * @return the name of the statevar the atom belongs to.
 */
inline std::string Atom::getStateVarName() {
  return stateVar->getName();
}
/**
 * @return the value of the atom.
 */
inline std::string Atom::getValue() {
  return value;
}
/**
 * @return the current label of the atom.
 */
inline std::string Atom::getLabel() {
  return label; 
}
/**
 * @return the type of the state var (Mode, Dependent, Observable, Command).
 */
inline std::string Atom::getVarType() {
  return stateVar->getType();
}
/**
 * @return the <Comp> the state var belongs to.
 */
inline Comp* Atom::getVarComp() {
  return stateVar->getComp();
}
/**
 * @return the references to premisses of transitions the atom belongs to.
 */
inline std::vector<Transition*>* Atom::getRefTransPre() {
  return ref_trans_pre;
}
/**
 * @return the references to post of transitions the atom belongs to.
 */
inline std::vector<Transition*>* Atom::getRefTransPost() {
  return ref_trans_post;
}
/**
 * @return the id of the atom, casted as a const char*.
 */
inline const char* Atom::getId() {
  return id.c_str();
}
/**
 * get another atom from the state variable this atom belongs to.
 * @param id a string that is the id of the searched atom.
 * @return an atom from the state variable.
 */
inline Atom* Atom::getAtomLink(const std::string &id) {
  for (unsigned int i=0; i<stateVar->getAtomLinks()->size(); i++)
    if (id == (*stateVar->getAtomLinks())[i]->getId())
      return (*stateVar->getAtomLinks())[i];
  return NULL;
}

};

#endif
