/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* Component */

#ifndef COMP_H
#define COMP_H

#include <string>
#include <vector>
#include "Transitionrt.h"
#include "Formula.h"
#include "StateVar.h"
#include "ModeVarrt.h"
#include "CondVarrt.h"
#include "ObsVarrt.h"
#include "DependentVarrt.h"
#include "CommandVar.h"
#include "../caen/caen-runt.h"
#include "../system/purge.h"

using namespace std;

#ifndef EQSTR_S
#define EQSTR_S
struct eqstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) == 0;
  }
};
#endif

/*class Atom;*/

/**
 * \class Comp
 * \brief A component is a transition system.
 */

class Comp {
 public:
  //type of the component.
  string type;
  //name of the component
  string name;
  /* number of atoms. */
  unsigned int atoms_number;
  //commands in.
  vector<CommandVar*>* cmd_in;
  //commands out.
  vector<CommandVar*>* cmd_out;
  //vector that stores the state variables that define the component
  vector<StateVar*>* state_variables;
  //vector of formulas that define the component's feasible states
  vector<Formula*>* constraints;
  vector<Clause*>* external_clauses; /**< external clauses (compiled constraints) */
  vector<variable>* variables; /**< continuous variables */
  vector<influence>* influences; /**< continuous influences */
  vector<Transition*>* component_transitions;  /**< transitions */
  vector<Comp*>* imported_comps; /**< imported components */
  //last transition.
  Transition* last_trans;
  hash_map<const char*, Atom*, hash<const char*>, eqstr>* atoms;
  double nominal_cover;
  string var_name;   /* parser/info variable */
  
 public:
  Comp();
  Comp(string na);
  Comp(string na, vector<StateVar*>* sta);
  Comp(string na, vector<CommandVar*> in, vector<CommandVar*>* out);
  ~Comp();

  //set the vector of state variables
  void setStateVariables(vector<StateVar*>* sta);
  //return the vector of state variables
  vector<StateVar*>* getStateVariables();
  //return a state variable, given its name.
  StateVar* getStateVariable(string na);
  //add a state variable
  void addStateVar(StateVar* sta_var);
  /* add a continuous variables. */
  void addVariable(variable va);
  //return all the atoms of the component.
  vector<Atom*>* getAllAtoms();
  //add an internal component.
  void addInternalComp(Comp* co);
  //compute nominal cover.
  void computeNominalCover();

  //BEHAVIOR RULES

  //set the vector of feasible states
  void setConstraints(vector<Formula*>* fea);
  //return the vector of feasible states
  vector<Formula*>* getConstraints();
  //add a Formula to feasible states
  void addFormulaToCS(Formula* frm);
  //add an influence.
  void addInfluence(influence inf);
  //set the vector of transitions
  void setCompTransitions(vector<Transition*>* trans);
  //return the vector of transitions
  vector<Transition*>* getCompTransitions();
  //add a Transition to the component's transitions
  void addTransitionToC(Transition* trans);

  //return the name of the component
  const char* getName();
  //set the name of the component
  void setName(string na);
  
  /**
   * Tools & debug.
   */
  void getTransitionsGuards();

  /**
   * debug & visu.
   */
  void showComp();

};

#endif
