/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Statert.h"
#include <iostream>

/* static variables. */
int State::state_created = 0;
int State::state_del = 0;

//constructors.
State::State() {
  state = new hash_map<const char*, TrajEl*, hash<const char*>, eqstr>();
#ifdef PROBAS
  cover = 0;
  trans_probs = new hash_map<int, double, hash<int> >();
  sorted_trans_probs = new multimap<double, int, greater_equal<double> >();
  temp_probs = new hash_map<const char*, hash_map<int, double, hash<int> >*, hash<const char*>, eqstr>();
#ifdef USE_BDD
  pb = NULL;
  bdd_state = NULL;
  no_obs = NULL;
  obs = NULL;
#endif
#endif
  comps = new hash_map<const char*, Comp*, hash<const char*>, eqstr>();
  rank = 0;
  st_lock = false;
#ifdef PROBAS
  obs_cond_prob = 1;
  prev_trans_prob = 0;
  proba = 0;
#endif
  atoms_number = 0;
  state_created++;
  observable = false;
  unknown = false;
#ifdef RECONF
  modifs = new vector<Atom*>();
#endif
}
/* 
 * make a state from the atoms' set.
 * Usually used as the first state of a trajectory.
 * The prevTrans are then initialized with id transitions consistent with the state.
 */
State::State(hash_map<const char*, Atom*, hash<const char*>, eqstr>* first_state) {
  TrajEl* elt;
  atoms_number = 0;
  state = new hash_map<const char*, TrajEl*, hash<const char*>, eqstr>();
  for (hash_map<const char*, Atom*, hash<const char*>, eqstr>::iterator it = first_state->begin(); it != first_state->end(); it++) {
    elt = new TrajEl((*it).second, NULL, NULL, (*it).second->getLabel());
    
    //debug
    /* std::cout << "created Elt:\n";
       elt->showTrajEl (true);
       std::cout << std::endl; */
    //debug

    if ((*it).second->support)
      {
	elt->support = (*it).second->support;  /* set the default support to the elt */
      
	//debug
	/* std::cout << "setting default support to: " << elt->getId () << ":\n";
	   (*it).second->support->showClause (true); */
	//debug
      }
    else elt->support = NULL;
    (*state)[(*it).second->getId()] = elt;
  }
  comps = new hash_map<const char*, Comp*, hash<const char*>, eqstr>();

  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it2 = state->begin(); 
       it2 != state->end(); it2++) { /* initializing prevTrans as id transitions and add components in comps. */
    if ((*it2).second->at->idle != NULL) {
      if (evalStateFormula((*it2).second->at->idle->pre, this)) { /* search for the id trans consistent with the state. */
	(*it2).second->prevTrans = (*it2).second->at->idle;
	atoms_number++;
	(*comps)[(*it2).second->at->getVarComp()->name.c_str()] = (*it2).second->at->getVarComp();
      }
    }
  }

#ifdef PROBAS
  computePrevTransProb();   /* set primary state probability. */
  proba = prev_trans_prob;
  cover = 0;
  trans_probs = new hash_map<int, double, hash<int> >();
  sorted_trans_probs = new multimap<double, int, greater_equal<double> >();
  temp_probs = new hash_map<const char*, hash_map<int, double, hash<int> >*, hash<const char*>, eqstr>();
#ifdef USE_BDD
  pb = NULL;
  bdd_state = NULL;
  no_obs = NULL;
  obs = NULL;
#endif
  obs_cond_prob = 1;
#endif
  rank = 0;
  st_lock = true;
  observable = false;
#ifdef RECONF
  modifs = NULL;
#endif
  unknown = false;
  state_created++;

  //debug
  //showFullState ();
  //debug
}

//kind of a copy constructor. Rewritten because of cycles of dependencies.
State::State(State* sta, bool locks) {
  TrajEl* el;
  state = new hash_map<const char*, TrajEl*, hash<const char*>, eqstr>();
  if (!locks) {
    st_lock = false;
    rank = 0;
#ifdef PROBAS
    obs_cond_prob = 1;
    prev_trans_prob = 0;
    proba = 0;
    cover = 0;
#ifdef USE_BDD
    obs = NULL;
    no_obs = NULL;
#endif
#endif
    atoms_number = 0;
    comps = new hash_map<const char*, Comp*, hash<const char*>, eqstr>();
    observable = 0;
  }
  else {
    st_lock = true;
    rank = sta->rank;
#ifdef PROBAS
    obs_cond_prob = sta->obs_cond_prob;
    prev_trans_prob = sta->prev_trans_prob;
    cover = sta->cover;
    proba = sta->proba;
#endif
    comps = new hash_map<const char*, Comp*, hash<const char*>, eqstr>(*sta->comps);
    atoms_number = sta->atoms_number;
    observable = sta->observable;
#ifdef USE_BDD
    if (sta->obs != NULL)
      obs = new bdd(*sta->obs);
    else no_obs = NULL;
    if (sta->no_obs != NULL)
      no_obs = new bdd(*sta->no_obs);
    else no_obs = NULL;
#endif
  }
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator st_it = sta->state->begin(); st_it != sta->state->end(); st_it++) {
    el = new TrajEl((*st_it).second);

    //debug
    /* std::cout << "from:\n";
       (*st_it).second->showTrajEl (true);
       std::cout << "created Elt2:\n";
       el->showTrajEl (true);
       std::cout << std::endl; */
    //debug

    if (!locks) {
      el->lock = 2;
      el->prevTrans = NULL;
    }
    (*state)[el->getId()] = el;
  }
  /* must loop again over the current state to update 'supported' set of elts :((. (bitcopy doesn't not seem to handle this).*/
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator st_it2 = sta->state->begin(); 
       st_it2 != sta->state->end(); st_it2++) {
    for (unsigned int i=0; i<(*st_it2).second->supported->size(); i++) {
      (*state)[(*st_it2).second->getId()]->supported->push_back((*state)[(*(*st_it2).second->supported)[i]->getId()]);
    }
  }
#ifdef PROBAS
  trans_probs = new hash_map<int, double, hash<int> >(*sta->trans_probs);
  sorted_trans_probs = new multimap<double, int, greater_equal<double> >(*sta->sorted_trans_probs);
  temp_probs = new hash_map<const char*, hash_map<int, double, hash<int> >*, hash<const char*>, eqstr>(*sta->temp_probs);
#ifdef USE_BDD
  if (sta->pb != NULL)
    pb = new bdd(*sta->pb);
  else pb = NULL;
  if (sta->bdd_state != NULL)
    bdd_state = new bdd(*sta->bdd_state);
  else bdd_state = NULL;
#endif
#endif
#ifdef RECONF
  if (sta->modifs != 0)
    modifs = new vector<Atom*>(*sta->modifs);
  else modifs = 0;
#endif
  unknown = sta->unknown;
  state_created++;
}

/**
 * Destructor for State.
 * The destructor has a rather unusual behavior due to its heavy use:
 * it does not destroy the set of elements the state is made of, because it can
 * be shared. To fully erase a state, consider first clearing the elements of the state
 * (a call to the class method clear ()), then delete it.
 */
State::~State() {
  delete comps;
  comps = 0;
#ifdef PROBAS
  delete trans_probs;
  trans_probs = 0;
  delete temp_probs;
  temp_probs = 0;
  delete sorted_trans_probs;
  sorted_trans_probs = 0;
#ifdef USE_BDD
  delete bdd_state;
  bdd_state = 0;
  delete pb;
  pb = 0;
  delete no_obs;
  no_obs = 0;
#endif
#endif
#ifdef RECONF
  delete modifs;
#endif
  state_del++;
}

/* set the prevTrans to an elt, and update transitions_cover. */
void State::setPrevTrans(const char* id, Transition* trans) {
  (*state)[id]->prevTrans = trans;
}

/* set the prevTrans to an elt (should be in state !!!), and update transitions_cover. */
void State::setPrevTrans(TrajEl* elt, Transition* trans) {
  elt->prevTrans = trans;
}

void State::lockStatePerm () {
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = state->begin ();
       it != state->end (); it++) {
    if ((*it).second->label == "true" && (*it).second->lock == 2) {
      for (unsigned int i=0; i<(*it).second->getStateVar()->getAtomLinks()->size(); i++) {
      if ((*(*it).second->getStateVar()->getAtomLinks())[i] != (*it).second->at) {
	(*state)[(*(*it).second->getStateVar()->getAtomLinks())[i]->getId()]->lock = 4;
      }
    }
    (*it).second->lock = 3;
    }
  }
  st_lock = true;
}

#ifdef PROBAS
/* return the posterior probability of the transition wr to the state, i.e. p(t/s). */
double State::getTransitionPostProb(Transition* trans) {
  if (trans_probs->count(trans->trans_id) != 0)
    return (*trans_probs)[trans->trans_id];
  else return 0;
}
#endif

/**
 * eval a stateformula considering a trajectory's temporal state. 
 * BEWARE : recursive formulas ARE NOT handled yet.
 * @param stfrm is the state formula to be evaluated.
 * @param sta is the state the evaluation is made in.
 * @result is true if the state formula is true, false otherwise.
 */
bool State::evalStateFormula(StateFormula* stfrm, State* sta) {
  bool resAtoms;
  bool tmp;
  multimap<bool, Atom*, eqbool>* sf_atoms;
  if (stfrm->operat == "and" | stfrm->operat == "") {
    if (stfrm->atoms!=NULL) {
      resAtoms=true;
      sf_atoms = stfrm->getAtoms();
      for (multimap<bool, Atom*, eqbool>::iterator i = sf_atoms->begin(); i != sf_atoms->end(); i++) {
	if ((*i).first)
	  tmp = (*sta->state)[(*i).second->getId()]->evalTrajEl();
	else tmp = !(*sta->state)[(*i).second->getId()]->evalTrajEl();
	resAtoms = resAtoms & tmp;
      }
      return resAtoms;
    }
  }
  else if (stfrm->operat=="or") {
    if (stfrm->atoms!=NULL) {
      resAtoms=false;
      sf_atoms = stfrm->getAtoms();
      for (multimap<bool, Atom*, eqbool>::iterator i = sf_atoms->begin(); i != sf_atoms->end(); i++) {
	if ((*i).first)
	  tmp = (*sta->state)[(*i).second->getId()]->evalTrajEl();
	else tmp = !(*sta->state)[(*i).second->getId()]->evalTrajEl();
	resAtoms = resAtoms | tmp;
      }
      return resAtoms;
    }
  }
  return 0;
}

/**
 * methods for probabilities computation.
 */
#ifdef PROBAS
/* compute the product of probabilities of transitions to that state. */
void State::computePrevTransProb() {
  double res = 1;
  /* in case there are some tCM. */
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = state->begin(); it != state->end(); it++) {
    if ((*it).second->prevTrans != NULL) {
      res = res * (*it).second->prevTrans->proba;
    }
  }
  //delete al_trans;
  prev_trans_prob = res;
}
/* compute the product of guarded probabilities of transitions to that state. */
double State::computePrevTransGuard() {
  double res = 1;
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = state->begin(); it != state->end(); it++) {
    if ((*it).second->prevTrans != NULL) {
      //res = res * (*it).second->prevTrans->proba_guard;
      res = res * getTransitionPostProb((*it).second->prevTrans);
    }
  }
  return res;
}
/* compute the product of guarded probabilities of transitions from that state. */
double State::computeNextTransGuard() {
  double res = 1;
  hash_map<int, int, hash<int> >* al_trans = new hash_map<int, int, hash<int> >();
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = state->begin(); it != state->end(); it++) {
    for (unsigned int i=0; i<(*it).second->nextTrans->size(); i++) {
      if (al_trans->count((*(*it).second->nextTrans)[i]->trans_id) == 0) {
	(*al_trans)[(*(*it).second->nextTrans)[i]->trans_id] = (*(*it).second->nextTrans)[i]->trans_id;
	res = res * (*(*it).second->nextTrans)[i]->proba_guard;
      }
    }
  }
  delete al_trans;
  return res;
}

#endif

/**
 * Debug & representation.
 */

/**
 * show the state in a 'light' way. 
 */
void State::showState() {
#ifdef PROBAS
  cout << "proba : " << proba << endl;
  cout << "cover : " << cover << endl;
#endif
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = state->begin(); it != state->end(); it++) {
    cout << (*it).first << " --" <<  (*it).second->label << " (" << (*it).second->propagation_number << ") " << " -- " << (*it).second->lock << endl;
 }
}
/**
 * show full state elts. 
 */
void State::showFullState() {
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = state->begin(); it != state->end(); it++) {
   (*it).second->showTrajEl(1);
 }
}
