/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Transitionrt.h"
#include <iostream>

/* Functions for class Transition. */

//constructors
Transition::Transition(StateFormula* pr, StateFormula* po, string ty) {
  pre = pr;
  post = po;
  type = ty;
  st_post = NULL;
  post_forms = new vector<Formula*>();
  idle = 0;
  modevar = NULL;
  addRefToAtoms();
  trans_id = 0;
  time_length = 1;
  weight = 0;
  commanded = true;
  mf = NULL;
  proba = cost = 0;
}
//with proba.
Transition::Transition(StateFormula* pr, StateFormula* po, string ty, double prob) {
  pre = pr;
  post = po;
  type = ty;
  st_post = NULL;
  post_forms = new vector<Formula*>();
  idle = 0;
  modevar = NULL;
  addRefToAtoms();
  proba = prob;
  weight = 0;
  trans_id = 0;
  time_length = 1;
  commanded = true;
  mf = NULL;
  cost = 0;
}
//with idle flag.
Transition::Transition(StateFormula* pr, StateFormula* po, string ty, double prob, bool idl) {
  pre = pr;
  post = po;
  type = ty;
  st_post = NULL;
  post_forms = new vector<Formula*>();
  idle = idl;
  modevar = NULL;
  addRefToAtoms();
  proba = prob;
  weight = 0;
  trans_id = 0;
  time_length = 1;
  commanded = true;
  mf = NULL;
  cost = 0;
}
Transition::Transition (StateFormula* pr, StateFormula* po, string ty, double prob, double co, bool idl) {
  pre = pr;
  post = po;
  type = ty;
  st_post = NULL;
  post_forms = new vector<Formula*>();
  idle = idl;
  modevar = NULL;
  addRefToAtoms();
  proba = prob;
  weight = 0;
  trans_id = 0;
  time_length = 1;
  commanded = true;
  mf = NULL;
  cost = co;
}
/* constructor for transition with multiple commands. */
Transition::Transition (StateFormula* pr, hash_map<int, StateFormula*, hash<int> >* s_post, string ty, double prob) {
  pre = pr;
  post = NULL;
  type = ty;
  st_post = s_post;
  post_forms = new vector<Formula*>();
  idle = 0;
  modevar = NULL;
  addRefToAtoms();
  proba = prob;
  weight = 0;
  trans_id = 0;
  time_length = st_post->size();
  commanded = true;
  mf = NULL;
}

//destructor.
Transition::~Transition() {
  delete pre;
  pre = 0;
  delete post;
  post = 0;
  purge(*post_forms);
  delete post_forms;
  delete mf;
}

/* set the id. */
void Transition::setID(int id) {
  trans_id = id;
}
//return the type of the transition
string Transition::getType() {
  return type;
}
//compare two transitions using the ids.
bool Transition::eq(Transition* tr) {
  if (trans_id == tr->trans_id) return true;
  else return false;
}
/* compilation : reference the transition to its atoms. */
void Transition::addRefToAtoms() {
  multimap<bool, Atom*, eqbool>* temp_pre;
  vector<Atom*>* temp_post;
  bool fail = false;
  if (type == "Fail")
    fail = true;
  if (pre != NULL) {
    temp_pre = pre->getAllSignedAtoms();
    for (multimap<bool, Atom*, eqbool>::iterator i=temp_pre->begin(); i != temp_pre->end(); i++) {  /* add ref to each atom. */
      refAtomPre ((*i).first, (*i).second);
    }
    delete temp_pre;
  }
  if (post != NULL) {
    temp_post = post->getAllAtoms();
    for (unsigned int j=0; j<temp_post->size(); j++)
      refAtomPost ((*temp_post)[j]);
    delete temp_post;
  }
  //add ref from formulas too.
  if (st_post != NULL) {
    for (hash_map<int, StateFormula*, hash<int> >::iterator st_it = st_post->begin(); st_it != st_post->end(); st_it++) {
      temp_post = (*st_it).second->getAllAtoms();
      for (unsigned int l=0; l<temp_post->size(); l++) {
	(*temp_post)[l]->addRefTransPost(this);
	(*temp_post)[l]->fail = fail;
	if (idle) (*temp_post)[l]->idle = this;
      }
      delete temp_post;
    }
  }
}

void Transition::refAtomPre (bool b, Atom* at) {
  if (b) {
    at->addRefTransPre(this);
    if (at->getVarType() == "Mode" & modevar == NULL) {  /* add reference to the state variable (Mode). */
      modevar = (ModeVar*) at->getStateVar();
      pre_at = at;   /* reference the pre atom. */
    }
  }
  else {  /* iterate through all others values of the state variable (negation). */
    vector<Atom*>* atoms = at->stateVar->atoms_links;
    for (unsigned int j=0; j<atoms->size(); j++)
      if (at != (*atoms)[j])
	(*atoms)[j]->addRefTransPre(this);
  }
}

void Transition::refAtomPost (Atom* at) {
  if (type == "Fail")
    at->fail = true;
  
  at->addRefTransPost(this);
  if (idle) at->idle = this;             /* set idle ref. */
  if (at->value == "unknown")
    unknown = 1;
  else unknown = 0;
  if (at->getStateVar() == modevar)      /* reference the post atom. */
    post_at = at;
}

//add a formula to the transition. Maybe formulas should be sorted ?
void Transition::addFormulaToTrans(Formula* form) {
  post_forms->push_back(form);
}

/* display */
void Transition::showMove() {
  if (!idle) {
    cout <<  pre_at->id << " --------> " << post_at->id << endl;
  }
}
//show the transition. !!! BEWARE : FORMULAS !!!
void Transition::showTransition() {
  cout << type << " (" << proba << ")";
  if (idle) cout << " (idle)";
  cout << " : ";
  if (pre!=NULL) pre->showStateFormula(); 
  cout << " =>O ";
  for (unsigned int i=0; i<post_forms->size(); i++) {
    cout << " and (";
    (*post_forms)[i]->showFormula();
    cout << ")";
  }
  if (post != NULL) {
    cout << " and ";
    post->showStateFormula();
  }
  else {
    for (hash_map<int, StateFormula*, hash<int> >::iterator it = st_post->begin(); it != st_post->end(); it++) {
      cout << "\n\t\t\t\t\t" << (*it).first << " : ";
      (*it).second->showStateFormula();
    }
    
  }
  cout << " -- " << trans_id << endl;
}
