/*
 * CHANGES:
 *
 * 01/11/01 : - creation (split for runtime).
 * 
 *  31/01/01 : - unitPropagate : move agenda from 'queue' to 'stack' for better results.
 *  18/12/00 : - added atoms post probabilities compilation from transitions
 *                 (compileTransitionsProb()).
 *  21/11/00 : - adding default clause (general one) to atoms during compilation.
 *  18/11/00 : - fixing bug on optimized getFirstUnit.
 */

#ifndef ITMS_H
#define ITMS_H

#include <vector>
#include <map>
#include <queue>
#include <time.h>
#include "../behav/Comprt.h"
#include "../behav/Clause.h"
#include "../behav/Atom.h"
#include "../behav/TrajElrt.h"
#include "../behav/Trajectoryrt.h"
#include "../system/purge.h"
#include "../system/config.h"
#include "../config.h"
#include "mdlChkrt.h"
#include "pi_table.h"
#include "math.h"

/**
 * \struct eqstr
 * Equality operator for const char* for hashtables (hash_map).
 */

#ifndef EQSTR_S
#define EQSTR_S
struct eqstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) == 0;
  }
};
#endif

/**
 * \class ItmsRes
 * \brief internal class for ITMS results.
 * It is made for storing useful links that reduce the search by focusing the conflicts detection.
 */

#ifndef ITMSRES_H
#define ITMSRES_H
class ItmsRes {
 public:
  
  State* res; /**< res is the working state. */
  hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* already_changed; /**< already_changed is the set of elements that already have been modified by the ITMS during the current search. */
  hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* chg_tmp; /**< subset of already_changed of elements recently moved. */
  vector<TrajEl*>* deleted; /**< elements that where recently deleted. */
  queue<Clause*>* forced_cstrs; /**< clauses that are false in the State res. */

  /**
   * Empty constructor.
   */
  ItmsRes() {
    res = new State();
    forced_cstrs = new queue<Clause*>();
    already_changed = new hash_map<const char*, TrajEl*, hash<const char*>, eqstr>();
    chg_tmp = new hash_map<const char*, TrajEl*, hash<const char*>, eqstr>();
    deleted = new vector<TrajEl*>();
  }
  /**
   * Constructor to wrap an existing State.
   */
  ItmsRes(State* wkst) {  
    res = wkst;
    forced_cstrs = new queue<Clause*>();
    already_changed = new hash_map<const char*, TrajEl*, hash<const char*>, eqstr>();
    chg_tmp = new hash_map<const char*, TrajEl*, hash<const char*>, eqstr>();
    deleted = new vector<TrajEl*>();
  }
  /**
   * Constructor that makes a wrap around a copy of an existing State.
   */
  ItmsRes(State* wkst, queue<Clause*>* fc) {
    res = new State(wkst, 1);
    forced_cstrs = new queue<Clause*>(*fc);
    chg_tmp = new hash_map<const char*, TrajEl*, hash<const char*>, eqstr>();
    deleted = new vector<TrajEl*>();
  }
  /**
   * Partial destructor. Full destruction is possible by making use of clear() before
   * deleting the object.
   */
  ~ItmsRes() {
    delete already_changed;
    already_changed = 0;
    delete chg_tmp;
    chg_tmp = 0;
    delete deleted;
    deleted = 0;
    delete forced_cstrs;
    forced_cstrs = 0;
  }
  /**
   * delete the State res and its elements, as already_changed.
   */
  void clear() { 
    res->clear();
    delete res;
    res = 0;
  }
  /**
   * test is the current wrapped state is empty.
   * @result true if empty, false otherwise.
   */
  bool resEmpty() {
    return res->state->empty();
  }
};
#endif

/**
 * ITMS (incremental truth maintenance system) based on William/Nayak 1997.
 * 
 * \class itms
 * \brief Incremental Truth Maintenance System.
 *
 */

class itms : public mdlChk {

 public:
  vector<Formula*>* constraints; /**< set of constraints made of formulas. It is deleted after model has been compiled into clauses.*/
  vector<StateVar*>* state_variables; /**< in order to compile all clauses at once... (internal + external). */
  bool lock_comp; /**< boolean flag for component insertion, as unlocked or locked (for observations). */
  PiTable* pit;  /* prime implicants */
  
  itms();
  itms(hash_map<const char*, Atom*, hash<const char*>, eqstr>* set, vector<Formula*>* cstrs, hash_map<const char*, Comp*, hash<const char*>, eqstr >* comps);
  ~itms();

  /* get an elt. */
  TrajEl* getElt(const char* id, State* wkst, Trajectory* tr, 
		 const int &timestep, const bool &trans_idle);
  TrajEl* getElt(const char* id, State* wkst, Trajectory* tr, 
		 const int &timestep);
  /*************************
   * Knowledge Compilation *
   *************************/
  void initialize();
  hash_map<const char*, Atom*, hash<const char*>, eqstr>* loadAtoms(vector<Clause*>* clauses);
  void idTransitions(Comp* co, int &curr_nu);
  void compileTransitionsProb(Atom* at);
  bool intersectClauses(vector<Clause*>* cls1, vector<Clause*>* cls2);
  bool intersectClauses(Atom* at, vector<Atom*>* vec);
  void rtInit ();
  void deadEndAtom (Atom* at);
#ifdef USE_BDD
  void initBddPb(State* wkst);
#else /* USE_BDD */
  void initPb (State* wkst);
#endif

  /*********************************
   * Truth Maintenance over States *
   *********************************/
  hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* copyState(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* state, State* wkst);
  int maxPropagationNumber(Clause* cla, TrajEl* p, State* wkst, Trajectory* tr, 
			   const int &timestep); //compute the max of the propagation_number of a clause.
  bool isResupportable(TrajEl* p, Clause* cla, const bool &non, State* wkst, Trajectory* tr, 
		       const int &timestep); //test if resupporting is possible.
  bool inVector(Atom* at, vector<Atom*>* vec); //test if at is in vec.
  bool inVector(const char* id, vector<const char*>* vec);
  bool inVector(TrajEl* at, vector<TrajEl*>* vec);
  bool inVector(Transition* tr, vector<Transition*>* vec);
  bool inVectorStateVar(StateVar* st, vector<StateVar*>* vec); //test if a state variable is in a vec.
  bool inVectorClause(Clause* cl, vector<Clause*>* vec);
  bool inMultimap(Atom* at, multimap<bool, Atom*, eqbool>* mul);
  hash_map<const char*, StateVar*, hash<const char*>, eqstr>* getStateVars(vector<TrajEl*>* vec); //return the set of statevars from the set of elts.
  void setSupport(Clause* cla, TrajEl* at, State* wkst, Trajectory* tr, 
		  const int &timestep);
  multimap<bool, TrajEl*, eqbool>* getNotUnknown(Clause* cla, State* wkst, Trajectory* tr, 
						 const int &timestep);
  TrajEl* getUnknown(Clause* cla, State* wkst, Trajectory* tr, const int &timestep);
  string evalAtomOnLabels(multimap<bool, Atom*, eqbool>::iterator it, State* wkst, Trajectory* tr, 
			  const int &timestep);
  bool testClause(Clause* cla, State* wkst, Trajectory* tr, const int &timestep);
  TrajEl* unitClause(Clause* cla, State* wkst, Trajectory* tr, const int &timestep);
  Clause* getFirstUnitClause(Atom* at, State* wkst, Trajectory* tr, const int &timestep); //return the first unit clause to be found in at's dependencies.
  Clause* getFirstUnitClause(State* wkst, Trajectory* tr, const int &timestep); //return the first unit clause from the entire set of clauses.
  Clause* getFirstUnitClause(vector<TrajEl*>* deleted, State* wkst, Trajectory* tr, 
			     const int &timestep); //more efficient...
  void resupportDependence(TrajEl* at, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* resc, State* wkst, Trajectory* tr, const int &timestep); //resupport labels.
  void checkForFocusedInternalConflicts(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* chg, queue<Clause*>* forced_cstrs, State* wkst, Trajectory* tr, const int &timestep);
  void checkForFocusedInternalConflicts(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* chg, vector<Clause*>* forced_cstrs, Trajectory* tr, const int &timestep);
  void checkForFocusedInternalConflictsCBFS(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* chg, vector<Clause*>* forced_cstrs, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* st);
  void checkForFocusedConflicts(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* chg, queue<Clause*>* forced_cstrs, State* wkst, Trajectory* tr, const int &timestep);
  Clause* checkForInternalConflicts(TrajEl* at, State* wkst, Trajectory* tr, const int &timestep);
  void checkForConflicts(vector<Clause*>* forced_cstrs, State* wkst, Trajectory* tr, const int &timestep);
  void propagateClause(Clause* cla, State* wkst, Trajectory* tr, const int &timestep);
  ItmsRes* propagateThroughConflict(Clause* cla, ItmsRes* curr_res, stack<ItmsRes*>*, 
				    const bool &nom, const double &rank, Trajectory* tr, 
				    const int &timestep); 
  void unitPropagate(ItmsRes* curr_st, vector<ItmsRes*>* res, 
		     const bool &nom, const double &rank, Trajectory* tr, 
		     const int &timestep);
  vector<ItmsRes*>* addPropositions(multimap<bool, TrajEl*, eqbool>* add_ats, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* already_changed, State* wkst, vector<ItmsRes*>* res, 
				    const bool &nom, 
				    const double &rank, const double &current_rank, 
				    Trajectory* tr, const int &timestep, 
				    const bool &comp_lock);
  vector<TrajEl*>* deleteClause(TrajEl* del_at, vector<TrajEl*>* res, 
				State* wkst, Trajectory* tr, const int &timestep);
  vector<TrajEl*>* traceSupport(TrajEl* p, vector<TrajEl*>* res, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* state, const bool &neg, const bool &flag); //trace the support of an atom.
  vector<TrajEl*>* traceTransitionsSupport(TrajEl* p, vector<TrajEl*>* res, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* state, const bool &neg, const bool &flag);
  vector<Atom*>* predictedAtoms(Atom* p, vector<Atom*>* res); //get the atoms that are predicted by an atom.
  double probaCondObsAtom(Atom* at, vector<Atom*>* obs); //return an estimation of the conditional probability of the atom given a set of observations.

  /****************************
   * Probabilities management *
   ****************************/
#ifdef ENTROPY
  double atomProba(Atom* at, State* wkst, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg);
  double entropy(Atom* at, State* wkst, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg);
  double entropy(StateVar* st, State* wkst, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg);
  double condProbTrans(Transition* trans, TrajEl* elt, State* wkst, double atom_prob, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg);
  void computeObsCondRough (State* wkst);
  double computeCondProbRough (TrajEl* elt, Transition* trans, State* wkst, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg);
  bool checkInOpposites (Atom* at, multimap<bool, Atom*, eqbool>* postt);
  bool checkInImportedComps (Atom* at, multimap<bool, Atom*, eqbool>* postt);
  void initStateProbas(State* wkst);
  void sortPostProbs(State* wkst);
#endif

  /*******************************************************
   * Trajectories management and consistency maintenance *
   *******************************************************/
  bool evalStateFormula(StateFormula* stfrm, Trajectory* tr, int timestep);

  int readPI ();

  /*************
   * debugging *
   *************/
  void showClause(Clause* cl);
  void viewAtomVector(vector<Atom*>* vec);
  void viewAtomVectorId(vector<Atom*>* vec);
  void viewClauseVector(vector<Clause*>* vec);
  void viewMultimapAtom(multimap<string, Atom*, eqstring>* mlt_map);
  void viewHashMapAtom(hash_map<const char*, Atom*, hash<const char*>, eqstr>* set);
  void viewTrajectoryConflicts(hash_map<int, vector<Clause*>*, hash<int> >* confs);
};

#endif
