#include "pi_gen.h"

Trajectory* initFakeTrajectory (itms* tms) {
  return new Trajectory (tms->active_atoms);
}

/* turn to 'unknown' the label of all atoms in the model */
Trajectory* setAtomsToUnknown (itms* tms) {
  Trajectory* tr = initFakeTrajectory (tms);
  State* fs = tr->getLastState ();
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = fs->state->begin ();
       it != fs->state->end (); it++) {
    (*it).second->label = "unknown";
  }
  return tr;
}

/* generate prime implicants for all propositions (atoms) in the qualitative model */
PiTable* generateAllPI (itms* tms) {
  PiTable* res = new PiTable ();
  Trajectory* tr_unk = setAtomsToUnknown (tms);
  Trajectory* cunk;
  vector<vector<Atom*>*>* pis;
  clock_t start, end;
  double elapsed;
  cout << "generating prime implicants... ";
  start = clock ();
  for (hash_map<const char*, Atom*, hash<const char*>, eqstr>::iterator it = tms->active_atoms->begin ();
       it != tms->active_atoms->end (); it++) {
    cout << "\nsearching " << (*it).second->getId () << "... \n";
    cunk = new Trajectory (tr_unk, 1);
    pis = generatePI ((*it).second, cunk, tms);
    res->addPIAtom ((*it).second, pis);
    res->showPI ((*it).second);
  }
  end = clock ();
  elapsed = ((double) (end - start))/CLOCKS_PER_SEC;
  cout << elapsed << " seconds.\n";
  delete tr_unk;
  return res;
}

/* inject an atom and get its prime implicants */
vector<vector<Atom*>*>* generatePI (Atom* at, Trajectory* tr, itms* tms) {
  vector<vector<Atom*>*>* res = new vector<vector<Atom*>*> ();
  vector<ItmsRes*>* tms_res;
  State* fs = tr->getLastState ();
  fs->st_lock = false;
  TrajEl* curr_el = tr->getElt (at->getId ());
  
  multimap<bool, TrajEl*, eqbool>* add_props = new multimap<bool, TrajEl*, eqbool> ();
  add_props->insert (pair<bool, TrajEl*> (1, curr_el)); /* insert a labeled 'true' element */
  
  tms_res = tms->addPropositions (add_props, NULL, fs, NULL, 0, 100, 0, tr, 0, 0);
  if (tms_res->empty ())
    cout << "Warning: no prime implicants found for " << at->getId () << endl;

  for (unsigned int i=0; i<tms_res->size (); i++) {
    res->push_back (new vector<Atom*> ());
    for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = (*tms_res)[i]->already_changed->begin (); it != (*tms_res)[i]->already_changed->end (); it++) {
      if ((*it).second->at != at && (*it).second->getLabel () == "true")
	(*res)[i]->push_back ((*it).second->at);
    }
    (*tms_res)[i]->clear ();
    delete (*tms_res)[i];
  }
  delete tms_res;

  return res;
}


/****************************************** enumeration method ******************************************************/

/**
 * enumerate all prime implicants for each atom (= proposition in the model).
 * Options (opt) are: - 1: enumerate all prime implicants. This can be highly computational.
 *                    - 2: enumerate prime implicants for nominal states only.
 */
void enumAllPI (itms* tms, int opt) {
  PiTable* res = new PiTable ();
  vector<vector<Atom*>*>* pis;
  clock_t start, end;
  double elapsed;
  Comp* curr_comp = NULL;
  int ok;

  /* check if prime implicants file is present and read it */
  if ((ok = tms->readPI(res)) == -1)
    cout << "error while reading prime implicants file\n";
  else if (ok == 1) {
    tms->pis = res;  /* set prime implicants */
    delete oco;  /* delete map of ranked components (from parser) */
    return;
  }

  tms->pis = res;  /* set prime implicants */
  Trajectory* tr_def = initFakeTrajectory (tms);
  cout << "generating prime implicants ";
  if (opt == 2)
    cout << "( nominal propositions only )";
  cout << " ...";
  start = clock ();
  for (map<int, component>::iterator comp_it = oco->begin ();
       comp_it != oco->end (); comp_it++) {
    curr_comp = (*comp_it).second->com;
    for (hash_map<const char*, Atom*, hash<const char*>, eqstr>::iterator it = curr_comp->atoms->begin ();
       it != curr_comp->atoms->end (); it++) {
      if (opt != 4 || !(*it).second->ref_trans_pre->empty ()) {
	if ((*it).second->getVarType () == "Dependent"
	    || (*it).second->getVarType () == "Observable"
	    /*|| (*it).second->getVarType () == "Cond"*/
	    || ((*it).second->getVarType () == "Command" 
		&& ((CommandVar*) (*it).second->getStateVar ())->out)) {
	  cout << "\nsearching " << (*it).second->getId () << "... ";
	  cout << endl;
	  pis = enumPI((*it).second, tr_def, tms, opt);
	  res->addPIAtom ((*it).second, pis);
	  /*res->showPI ((*it).second);*/
	}
      }
    }
  }
  end = clock ();
  elapsed = ((double) (end - start))/CLOCKS_PER_SEC;
  cout << elapsed << " seconds.\n";

  /* writing file of prime implicants. */
  if (!ok)  res->writePI(appliName);

  delete tr_def;
  delete oco;  /* delete map of ranked components (from parser) */
}

vector<vector<Atom*>*>* enumPI (Atom* at, Trajectory* tr, itms* tms, int opt) {
  unsigned int lim, inc;
  vector<Atom*>* single;
  multimap<int, vector<Atom*>*>* pi2test = new multimap<int, vector<Atom*>*> ();
  vector<vector<Atom*>*>* res = new vector<vector<Atom*>*> ();
  Comp* atc = at->getStateVar ()->comp_lk;
  vector<Comp*>* comps = getImportedComps (atc, NULL);
  comps->push_back (atc);
  unsigned int comp_size = comps->size ();

  single = enumSingletons (comps, pi2test, at, tr, tms, res, opt);
  lim = single->size ();
  
  if (opt != 3) {
    inc=1;
    while (inc < lim) {
      inc++;
      if (! enumNextSize (at, single, pi2test, inc, tr, tms, res, comp_size))
	break;
    }
    for (multimap<int, vector<Atom*>*>::iterator it = pi2test->begin ();
	 it != pi2test->end (); it++)
      delete (*it).second;
    delete pi2test;
  }
  else {
    /*cout << "single size: " << single->size () << endl;*/
    enumCompleteSize (at, single,  tr, tms, res);
  }
  
  delete comps;
  delete single;
  
  if (res->empty ())
    cout << "Warning: no prime implicants found for " << at->getId () << endl;
  
  return res;
}

vector<Atom*>* enumSingletons (vector<Comp*>* comps, multimap<int, vector<Atom*>*>* pi2test, Atom* at, Trajectory* tr, itms* tms, vector<vector<Atom*>*>* res, int opt) {
  vector<Atom*>* vec_tmp;
  vector<Atom*>* rvec = new vector<Atom*> ();
  Comp* cmp;
  StateVar* stt;
  
  cout << "\t( " << comps->size () << " comp )\n";
  for (unsigned int i=0; i<comps->size (); i++) {
    cmp = (*comps)[i];
    for (unsigned int i=0; i<cmp->state_variables->size (); i++) {
      stt = (*cmp->state_variables)[i];
      if (stt->getType () == "Mode"
	  || (stt->getType () == "Command" && ! ((CommandVar*) stt)->out)
	  || stt->getType () == "Cond") {
	for (unsigned int j=0; j<stt->atoms_links->size (); j++) {
	  if (opt != 2 || opt != 3 || ! (*stt->atoms_links)[j]->fail) {
	    if (stt->getType () != "Mode" || (*stt->atoms_links)[j]->value != "unknown") {
	      vec_tmp = new vector<Atom*> ();
	      vec_tmp->push_back ((*stt->atoms_links)[j]);
	      if (opt != 3 || comps->size () <= 5) {
		if (satisTms (vec_tmp, at, tr, tms, res)) {
		  vec_tmp =  new vector<Atom*> ();
		  vec_tmp->push_back ((*stt->atoms_links)[j]);
		  pi2test->insert (pair<int, vector<Atom*>* >(1, vec_tmp));
		  rvec->push_back ((*stt->atoms_links)[j]);
		}
	      }
	      else rvec->push_back ((*stt->atoms_links)[j]);
	    }
	  }
	}
      }
    }
  }
  /*cout << "( " << rvec->size () << " )\n";*/
  return rvec;
}

bool enumNextSize (Atom* current_at, vector<Atom*>* single, multimap<int, vector<Atom*>*>* pi2test, unsigned int set_size, Trajectory* tr, itms* tms, vector<vector<Atom*>*>* res, unsigned int comp_size) {
  bool flag = false;
  Atom* at_sing;
  vector<Atom*>* vec_tmp;
  vector<Atom*>* vec_sing = new vector<Atom*> ();
  pair<multimap<int, vector<Atom*>*>::iterator, multimap<int, vector<Atom*>*>::iterator> mult_set;
  multimap<int, vector<Atom*>*>::iterator it_mult;

  for (unsigned int i=0; i<single->size (); i++) {
    at_sing = (*single)[i];
    vec_sing->push_back (at_sing);
    mult_set = pi2test->equal_range (set_size-1);
    it_mult = mult_set.first;
    while (it_mult != mult_set.second) {
      /*if (! inVector (vec_sing, (*it_mult).second)) {*/
      if (! hasStVar ((*it_mult).second, at_sing, vec_sing)) {
	/*if (set_size < comp_size) {
	  vec_tmp = new vector<Atom*> (*(*it_mult).second);
	  vec_tmp->push_back (at_sing);
	  pi2test->insert (pair<int, vector<Atom*>*> (set_size, vec_tmp));
	  flag = true;
	}
	else*/ if (insertAtomInVec (current_at, at_sing, (*it_mult).second, tr, tms, res)) {
	  vec_tmp = new vector<Atom*> (*(*it_mult).second);
	  vec_tmp->push_back (at_sing);
	  pi2test->insert (pair<int, vector<Atom*>*> (set_size, vec_tmp));
	  flag = true;
	}
      }
      it_mult++;
    }
  }
  delete vec_sing;
  return flag;
}

void enumCompleteSize (Atom* current_at, vector<Atom*>* single, Trajectory* tr, itms* tms, vector<vector<Atom*>*>* res) {
  bool completed = false;
  vector<Atom*>* vec;
  vector<vector<Atom*>*>* candis = new vector<vector<Atom*>*> ();
  unsigned int num = 0;

  for (unsigned int i=0; i<single->size (); i++) {
    vec = new vector<Atom*> ();
    vec->push_back ((*single)[i]);
    candis->push_back (vec);
  }

  /* construct the sets of candidates */
  cout << "single size: " << single->size () << endl;
  while (! completed) {
    completed = cartVec (candis, single, num++);
    cout << "candis size: " << candis->size () << endl;
  }

  /* test the sets */
  cout << "candidates size: " << candis->size () << endl;
  for (unsigned int i=1; i<candis->size (); i++) {
    satisTms ((*candis)[i], current_at, tr, tms, res);
  }
  
  delete candis;
}

bool cartVec (vector<vector<Atom*>*>* candis, vector<Atom*>* single, unsigned int num) {
  vector<Atom*> *curr, *vec_tmp1, *vec_tmp2, *vec_store;
  vector<vector<Atom*>*>* res = new vector<vector<Atom*>*> ();
  Atom* at;
  bool add = false;
  bool flag = false;
  
  for (unsigned int i=0; i<candis->size (); i++) {
    curr = (*candis)[i];
    if (curr->size () >= num) {
      add = false;
      vec_tmp1 = new vector<Atom*> (*(*candis)[i]);
      vec_store = new vector<Atom*> ();
      for (unsigned int j=0; j<single->size (); j++) {
	at = (*single)[j];
	vec_store->push_back (at);
	if (! hasStVar (vec_tmp1, at, vec_store)) {
	  if (! add) {
	    add = true;
	    flag = true;
	    curr->push_back (at);
	  }
	  else {
	    vec_tmp2 = new vector<Atom*> (*vec_tmp1);
	    vec_tmp2->push_back (at);
	    res->push_back (vec_tmp2);
	  }
	}
      }
      delete vec_store;
      delete vec_tmp1;
    }
  }
  /*cout << "res size: " << res->size () << endl;*/
  if (res->empty () && !flag) {
    delete res;
    return true;
  }
  for (unsigned int i=0; i<res->size (); i++) {
    candis->push_back ((*res)[i]);
  }
  delete res;
  return false;
}

bool hasStVar (vector<Atom*>* candis, Atom* at, vector<Atom*>* store) {
  for (unsigned int i=0; i<candis->size (); i++) {
    if ((*candis)[i] == at
	|| at->getStateVar () == (*candis)[i]->getStateVar ()
	|| inVector (store, (*candis)[i])) {
      return true;
    }
  }
  return false;
}

bool insertAtomInVec (Atom* current_at, Atom* at, vector<Atom*>* vec, Trajectory* tr, itms* tms, vector<vector<Atom*>*>* res) {
  vector<Atom*>* vec_tmp = new vector<Atom*> (*vec);
  vec_tmp->push_back (at);
  if (! inImplicants (vec_tmp, tms, res)) {
    if (satisTms (vec_tmp, current_at, tr, tms, res)) {
      delete vec_tmp;
      return true;
    }
  }
  return false;
}

bool inImplicants (vector<Atom*>* vec, itms* tms, vector<vector<Atom*>*>* res) {
  bool bres, bat;
  for (unsigned int i=0; i<res->size (); i++) {
    bres = true;
    for (unsigned int j=0; j<(*res)[i]->size (); j++) {
      bat = false;
      for (unsigned int k=0; k<vec->size (); k++) {
	if ((*(*res)[i])[j] == (*vec)[k])
	  bat = true;
      }
      bres = bres & bat;
      if (! bres)
	break;
    }
    if (bres)
      return bres;
  }
  return false;
}

vector<Comp*>* getImportedComps (Comp* co, vector<Comp*>* res) {
  if (res == NULL) res = new vector<Comp*> ();
  if (co->imported_comps == NULL)
    return res;
  for (unsigned int i=0; i<co->imported_comps->size (); i++) {
    if (! inVector ((*co->imported_comps)[i], res)) {
      res->push_back ((*co->imported_comps)[i]);
      res = getImportedComps ((*co->imported_comps)[i], res);
    }
  }
  return res;
}

bool inVector (Comp* co, vector<Comp*>* comps) {
  for (unsigned int i=0; i<comps->size (); i++)
    if ((*comps)[i] == co)
      return true;
  return false;
}

bool inVector (vector<Atom*>* at, vector<Atom*>* vec) {
  for (unsigned int i=0; i<vec->size (); i++)
    for (unsigned int j=0; j<at->size (); j++)
      if ((*vec)[i] == (*at)[j])
	return true;
  return false;
}

bool inVector (vector<Atom*>* vec, Atom* at) {
  for (unsigned int i=0; i<vec->size (); i++)
    if ((*vec)[i] == at)
      return true;
  return false;
}

bool satisTms (vector<Atom*>* props, Atom* current_at, Trajectory* tr, itms* tms, vector<vector<Atom*>*>* res) {
  vector<ItmsRes*>* tms_res;
  Trajectory* ctr = new Trajectory (tr, 1);
  State* fs = ctr->getLastState ();
  fs->st_lock = false;
  multimap<bool, TrajEl*, eqbool>* add_tmp = new multimap<bool, TrajEl*, eqbool> ();

  add_tmp->insert (pair<bool, TrajEl*> (1, ctr->getElt (current_at->getId ())));  /* positive cyle */
  
  for (unsigned int i=0; i<props->size (); i++)
    add_tmp->insert (pair<bool, TrajEl*> (1, ctr->getElt ((*props)[i]->getId ())));
  tms_res = tms->testImplicants (add_tmp, NULL, fs, NULL, 0, 100, 0, ctr, 0, 0, 1);
  add_tmp->clear ();
  delete ctr->traj;
  if (tms_res->empty ()) {
    delete props;
    delete tms_res;
    return false;  /* no solution. get rid of the branch */
  }
  for (unsigned int j=0; j<tms_res->size (); j++) {
    (*tms_res)[j]->clear ();
    delete (*tms_res)[j];
  }
  delete tms_res;
  
  ctr = new Trajectory (tr, 1);  /* negative cycle */
  fs = ctr->getLastState ();
  fs->st_lock = false;
  add_tmp->insert (pair<bool, TrajEl*> (0, ctr->getElt (current_at->getId ())));
  for (unsigned int i=0; i<props->size (); i++)
    add_tmp->insert (pair<bool, TrajEl*> (1, ctr->getElt ((*props)[i]->getId ())));
  tms_res = tms->testImplicants (add_tmp, NULL, fs, NULL, 0, 100, 0, ctr, 0, 0, 0);
  delete add_tmp;
  if (tms_res->empty ()) {
    cout << "\t=> ";
    for (unsigned int j=0; j<props->size (); j++)
      cout << (*props)[j]->getId () << " /\\ ";
    cout << endl;
    res->push_back (props);
    delete ctr->traj;
  }
  else {
    for (unsigned int j=0; j<tms_res->size (); j++) {
      (*tms_res)[j]->clear ();
      delete (*tms_res)[j];
    }
    delete tms_res;
    delete ctr->traj;
    return true;  /* all right but not a prime */
  }
  delete tms_res;
  return false;  /* a prime has been found */
}
