/**
 * Modified 08/2002 (rainy summer...) Emmanuel Benazera 
 *
 * - fixes and warnings,
 * - bounded results.
 * 
 **/

/*********************************************************************/
/*                                                                   */
/*                          --- nelder.c ---                         */
/*                                                                   */
/* fct: module des procedures d'optimisation non-lineaire avec le    */
/*      simplexe de Nelder & Mead.                                   */
/*                                                                   */
/* aut: Francois Dufour (LAAS)                                       */
/*                                                                   */
/* ver: 2 mars 1994                                                  */
/*                                                                   */
/*********************************************************************/

#include <stdlib.h>
#include <math.h>
#include <stdio.h>

#include "nelder.h"


/*********************************************************************/
/*                                                                   */
/*                      --- Nelder_initial ---                       */
/*                                                                   */
/* fct: initialise le simplexe.                                      */
/*                                                                   */
/* ver: 1 septembre 1993                                             */
/*                                                                   */
/*********************************************************************/

void Nelder_initial
(
int     type_init,   /* type d'initialisation du simplexe         */
double  param_init,  /* parametre d'initialisation                */
double *Xi,          /* ptr du point initial                      */
double *Xmin,        /* ptr des bornes minimales du domaine des X */
double *Xmax,        /* ptr des bornes maximales du domaine des X */
Nelder_simplexe *simplexe      /* ptr du simplexe                 */
)
{
   double delta1, delta2;
   int  i, j, init_ran0;

   init_ran0 = 0;
   for (j = 0; j < simplexe->pt[0]->n; j++)
      simplexe->pt[0]->x[j] = Xi[j];

   switch (type_init)
   {
   case 1:

   /*** simplexe regulier ***/

      delta1 = param_init*(sqrt(simplexe->pt[0]->n+1.0) + simplexe->pt[0]->n-1.0)/simplexe->pt[0]->n;
      delta2 = param_init*(sqrt(simplexe->pt[0]->n+1.0) - 1.0)/simplexe->pt[0]->n;
      for (i = 1; i < simplexe->dim; i++)
         {
         for (j = 0; j < simplexe->pt[i]->n; j++)
            {
            if (i == (j+1))
               simplexe->pt[i]->x[j] = Xi[j] + delta1*(Xmax[j] - Xmin[j]);
            else
               simplexe->pt[i]->x[j] = Xi[j] + delta2*(Xmax[j] - Xmin[j]);
            }
         }
   break;

   case 2:

   /*** simplexe orthogonal ***/ 

      for (i = 1; i < simplexe->dim; i++)
         {
         for (j = 0; j < simplexe->pt[i]->n; j++)
            {
            if (i == (j+1))
               simplexe->pt[i]->x[j] = Xi[j] + param_init*(Xmax[j] - Xmin[j]);
            else
               simplexe->pt[i]->x[j] = Xi[j];
            }
         }
   break;

   case 3:

   /*** simplexe aleatoire ***/

      for (i = 1; i < simplexe->dim; i++)
         {
         for (j = 0; j < simplexe->pt[i]->n; j++)
            simplexe->pt[i]->x[j] = Xmin[j] + Optim_ran0(&init_ran0)*(Xmax[j] - Xmin[j]);
         }
   break;

   case 4:

   /*** simplexe aleatoire centre ***/

      for (i = 1; i < simplexe->dim; i++)
         {
         for (j = 0; j < simplexe->pt[i]->n; j++)
            simplexe->pt[i]->x[j] = Xi[j] + param_init*(Xmax[j]-Xmin[j])*(Optim_ran0(&init_ran0)-0.5);
         }
   break;
   }
}


/*********************************************************************/
/*                                                                   */
/*                        --- Nelder_pivot ---                       */
/*                                                                   */
/* fct: calcul le point pivot du simplexe.                           */
/*      (utilise pour le calcul de la direction de deplacement)      */
/*                                                                   */
/* ver: 10 juin 1993                                                 */
/*                                                                   */
/*********************************************************************/

void Nelder_pivot
(
int              type_pivot,   /* type de pivot */
Nelder_simplexe *simplexe,     /* ptr du simplexe */
Nelder_point    *pivot         /* ptr du point pivot */
)
{
   int  i, j;
   double somme;

   switch (type_pivot)
   {
   case 1:

   /*** moyenne arithmetique de tous les pts sauf fmax ***/

      for (j = 0; j < pivot->n; j++)
         pivot->x[j] = 0.0;

      for (i = 0; i < simplexe->dim; i++)
         {
         if (i != simplexe->max)
            {
            for (j = 0; j < pivot->n; j++)
               pivot->x[j] += simplexe->pt[i]->x[j];
            }
         }
      for (j = 0; j < pivot->n; j++)
         pivot->x[j] /= pivot->n;
   break;

   case 2:

   /*** moyenne ponderee de tous les pts sauf fmax ***/

      for (j = 0; j < pivot->n; j++)
         pivot->x[j] = 0.0;

      somme = 0.0;
      for (i = 0; i < simplexe->dim; i++)
         {
         if (i != simplexe->max)
            {
            somme += simplexe->pt[simplexe->max]->fx - simplexe->pt[i]->fx;
            for (j = 0; j < pivot->n; j++)
               pivot->x[j] += (simplexe->pt[simplexe->max]->fx - simplexe->pt[i]->fx) * simplexe->pt[i]->x[j];
            }
         }
      for (j = 0; j < pivot->n; j++)
         pivot->x[j] /= somme;
   break;

   case 3:

   /*** fmin ***/

      for (j = 0; j < pivot->n; j++)
         pivot->x[j] = simplexe->pt[simplexe->min]->x[j];
   break;
   }
}


/*********************************************************************/
/*                                                                   */
/*                     --- Nelder_tri_simplexe ---                   */
/*                                                                   */
/* fct: effectue le tri des points du simplexe suivant un cout       */
/*      ascendant.                                                   */
/*                                                                   */
/* not: algorithme appele "straight insertion"                       */
/*      ("Numerical Recipes", page 242)                              */
/*                                                                   */
/* car: index de fx_min = 0                                          */
/*      index de fx_max = (n+1) - 1                                  */
/*                                                                   */
/* ver: 14 fevrier 1994                                              */
/*                                                                   */
/*********************************************************************/

void Nelder_tri_simplexe
(
Nelder_simplexe *simplexe   /* ptr du simplexe */
)
{
   int  i, j;
   Nelder_point *pt_a_trier;

   for (j = 1; j < simplexe->dim; j++)
      {
      pt_a_trier = simplexe->pt[j];
      i = j-1;
      while ((i >= 0) && (simplexe->pt[i]->fx > pt_a_trier->fx))
         {
         simplexe->pt[i+1] = simplexe->pt[i];
         i--;
         }
      simplexe->pt[i+1] = pt_a_trier;
      }
}


/*********************************************************************/
/*                                                                   */
/*                       --- Nelder_reflexion ---                    */
/*                                                                   */
/* fct: calcul de la reflexion du simplexe.                          */
/*                                                                   */
/* ver: 10 juin 1993                                                 */
/*                                                                   */
/*********************************************************************/

void Nelder_reflexion
(
double        alpha,      /* facteur de reflexion */
Nelder_point *pivot,      /* ptr du point pivot */
Nelder_point *fx_max,     /* ptr du point du simplexe avec fx_max */
Nelder_point *reflexion,  /* ptr du point resultant de la reflexion */
double* Xmin,
double* Xmax
)
{
   int  i;

   for (i = 0; i < reflexion->n; i++) {
      reflexion->x[i] = (1.0 + alpha) * pivot->x[i] - alpha * fx_max->x[i];
      if (reflexion->x[i] > Xmax[i]) reflexion->x[i] = Xmax[i];
      else if (reflexion->x[i] < Xmin[i]) reflexion->x[i] = Xmin[i];
   }
}


/*********************************************************************/
/*                                                                   */
/*                       --- Nelder_expansion ---                    */
/*                                                                   */
/* fct: calcul de l'expansion du simplexe.                           */
/*                                                                   */
/* ver: 10 juin 1993                                                 */
/*                                                                   */
/*********************************************************************/

void Nelder_expansion
(
double        gamma,       /* facteur d'expansion */
Nelder_point *pivot,       /* ptr du point pivot */
Nelder_point *reflexion,   /* ptr du point resultant de la reflexion */
Nelder_point *expansion,   /* ptr du point resultant de l'expansion */
double* Xmin,
double* Xmax
)
{
   int  i;

   for (i = 0; i < expansion->n; i++) {
     expansion->x[i] = gamma * reflexion->x[i] + (1.0 - gamma) * pivot->x[i];
     if (expansion->x[i] > Xmax[i]) expansion->x[i] = Xmax[i];
     else if (expansion->x[i] < Xmin[i]) expansion->x[i] = Xmin[i];
   }
}


/*********************************************************************/
/*                                                                   */
/*                      --- Nelder_contraction ---                   */
/*                                                                   */
/* fct: calcul de la contraction du simplexe.                        */
/*                                                                   */
/* ver: 10 juin 1993                                                 */
/*                                                                   */
/*********************************************************************/

void Nelder_contraction
(
double        beta,         /* facteur de contraction */
Nelder_point *pivot,        /* ptr du point pivot */
Nelder_point *fx_max,       /* ptr du point du simplexe avec fx_max */
Nelder_point *contraction,  /* ptr du point resultant de la contraction */
double* Xmin,
double* Xmax
)
{
   int  i;

   for (i = 0; i < contraction->n; i++) {
     contraction->x[i] = beta * fx_max->x[i] + (1.0 - beta) * pivot->x[i];
     if (contraction->x[i] > Xmax[i]) contraction->x[i] = Xmax[i];
     else if (contraction->x[i] < Xmin[i]) contraction->x[i] = Xmin[i];
   }
}


/*********************************************************************/
/*                                                                   */
/*                       --- Nelder_reduction ---                    */
/*                                                                   */
/* fct: calcul de la reduction du simplexe (reduction a partir du    */
/*      point fmin).                                                 */
/*                                                                   */
/* ver: 30 aout 1993                                                 */
/*                                                                   */
/*********************************************************************/

void Nelder_reduction
(
double           delta,     /* facteur de reduction */
Nelder_simplexe *simplexe   /* ptr du simplexe */
)
{
   int  i, j;

   for (i = 0; i < simplexe->dim; i++)
      {
      if (i != simplexe->min)
         {
         for (j = 0; j < simplexe->pt[i]->n; j++)
            simplexe->pt[i]->x[j] = (1.0/delta)*simplexe->pt[i]->x[j]
                 + (1.0 - (1.0/delta))*simplexe->pt[simplexe->min]->x[j];
         }
      }
}


/*********************************************************************/
/*                                                                   */
/*                    --- Nelder_critere_arret ---                   */
/*                                                                   */
/* fct: evaluation du critere d'arret du simplexe.                   */
/*                                                                   */
/* ver: 14 fevrier 1994                                              */
/*                                                                   */
/*********************************************************************/

boolean Nelder_critere_arret
(
int     type_critere_arret,   /* type de critere d'arret */
double  precision_cout,       /* precision recherchee sur le cout */
double  precision_X,          /* precision recherchee sur les X */
double *delta_X,              /* ptr des largeurs des domaines */
Nelder_simplexe *simplexe     /* ptr du simplexe */
)
{
   int  i, j;
   boolean arret;
   double erreur, moyenne, temp;
   double *Xmoyen;

   arret = FALSE;

   switch (type_critere_arret)
      {
      case 1:   

      /*** moyenne_ecarts_couts < precision_cout (critere de Nelder) ***/

         moyenne = 0.0;
         for (i = 0; i < simplexe->dim; i++)
            moyenne += simplexe->pt[i]->fx / simplexe->dim;

         erreur = 0.0;
         for (i = 0; i < simplexe->dim; i++)
            {
            temp = simplexe->pt[i]->fx - moyenne;
            erreur += temp * temp;
            }
         erreur = sqrt(erreur / simplexe->dim);

         if (erreur < precision_cout)
            arret = TRUE;
      break;

      case 2:   

      /*** |ecarts_couts| < precision_cout ***/

         moyenne = 0.0;
         for (i = 0; i < simplexe->dim; i++)
            moyenne += simplexe->pt[i]->fx / simplexe->dim;

         arret = TRUE;
         i = 0;
         while ((arret) && (i < simplexe->dim))
            {
            if (abs(simplexe->pt[i]->fx - moyenne) > precision_cout)
               arret = FALSE;
            i++;
            }
      break;

      case 3:   

      /*** (Xi - Xmoyen) < precision_X ***/

         Xmoyen = (double *) malloc(simplexe->pt[0]->n * sizeof(double));

         for (j = 0; j < simplexe->pt[0]->n; j++)
            {
            Xmoyen[j] = 0.0;
            for (i = 0; i < simplexe->dim; i++)
               Xmoyen[j] += simplexe->pt[i]->x[j];
            Xmoyen[j] /= simplexe->dim;
            }

         arret = TRUE;
         i = 0;
         while ((arret) && (i < simplexe->dim))
            {
            if (Optim_longvec_norm(simplexe->pt[0]->n, Xmoyen, simplexe->pt[i]->x, delta_X) > precision_X)
               arret = FALSE;
            i++;
            }

         free(Xmoyen);
      break;
      }

   return arret;
}


/*********************************************************************/
/*                                                                   */
/*                     --- Nelder_trace_initial ---                  */
/*                                                                   */
/* fct: ecriture initiale des donnees du simplexe.                   */
/*                                                                   */
/* ver: 12 decembre 1993                                             */
/*                                                                   */
/*********************************************************************/

void Nelder_trace_initial
(
FILE *fich_trace,           /* flux de sortie */
Nelder_simplexe *simplexe   /* ptr du simplexe */
)
{
  int j;

   for (j = 0; j < simplexe->pt[0]->n; j++)
      fprintf(fich_trace,"============  ");
   fprintf(fich_trace,"============ \n");
   for (j = 1; j <= simplexe->pt[0]->n; j++)
      fprintf(fich_trace,"     x%-4d    ", j);
   fprintf(fich_trace,"    cout \n");
   for (j = 0; j < simplexe->pt[0]->n; j++)
      fprintf(fich_trace,"============  ");
   fprintf(fich_trace,"============ \n\n");
}


/*********************************************************************/
/*                                                                   */
/*                        --- Nelder_trace ---                       */
/*                                                                   */
/* fct: ecriture des donnees du simplexe.                            */
/*                                                                   */
/* ver: 10 juin 1993                                                 */
/*                                                                   */
/*********************************************************************/

void Nelder_trace
(
FILE *fich_trace,           /* flux de sortie */
Nelder_simplexe *simplexe   /* ptr du simplexe */
)
{
   int i, j;

   for (i = simplexe->dim-1; i >= 0; i--)
      {
      for (j = 0; j < simplexe->pt[i]->n; j++)
         fprintf(fich_trace,"%12.6g  ", simplexe->pt[i]->x[j]);
      fprintf(fich_trace,"%12.6g \n", simplexe->pt[i]->fx);
      }
   fprintf(fich_trace,"\n");
}


/*********************************************************************/
/*                                                                   */
/*                    --- Nelder_trace_final ---                     */
/*                                                                   */
/* fct: ecriture des donnees sur le calcul du simplexe.              */
/*                                                                   */
/* ver: 17 juin 1993                                                 */
/*                                                                   */
/*********************************************************************/

void Nelder_trace_final
(
FILE *fich_trace,         /* flux de sortie */
int   etat_calcul,        /* etat des calculs */
int   nb_iter_simplexe,   /* nb iteration dans l'algorithme du simplexe */
int   nb_eval_cout        /* nb evaluation de la fonction cout */
)
{
   fprintf(fich_trace,"Etat des calculs: ");
   if (etat_calcul == MINIMUM_ATTEINT)
      fprintf(fich_trace,"MINIMUM ATTEINT\n");
   else if (etat_calcul == ERREUR_MAX_EVAL_COUT)
      fprintf(fich_trace,"NOMBRE MAXIMUM D'EVALUATIONS DU COUT ATTEINT\n");
   else if (etat_calcul == ERREUR_CALCUL_COUT)
      fprintf(fich_trace,"ERREUR DANS LE CALCUL DU COUT\n");
   else
      fprintf(fich_trace,"CAS INCONNU\n");

   fprintf(fich_trace,"Nombre d'iterations du simplexe: %8d\n", nb_iter_simplexe);
   fprintf(fich_trace,"Nombre d'evaluations du cout:    %8d\n\n", nb_eval_cout);
}


/*********************************************************************/
/*                                                                   */
/*                     --- Nelder_init_options ---                   */
/*                                                                   */
/* fct: initialise les options du simplexe en lisant le fichier:     */
/*      "nelder.dat".                                                */
/*                                                                   */
/* ver: 17 fevrier 1994                                              */
/*                                                                   */
/*********************************************************************/

void Nelder_init_options
(
Nelder_options *options   /* ptr du vecteur des options */
)
{
   FILE *entree;
   int   i;

   entree = fopen("nelder.dat","r");

   if (entree)
   {
   for (i = 1; i <= 6; i++)
      {fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");}

   fscanf(entree, "%d", &(options->TRACE));
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");
   fscanf(entree, "%d", &(options->TRACE_FICHIER));
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");
   fscanf(entree, "%s", &(options->fichier_trace));
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");

   for (i = 1; i <= 3; i++)
      {fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");}
   fscanf(entree, "%d", &(options->TRACE_OPT));
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");
   fscanf(entree, "%d", &(options->FICHIER_OPT));
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");
   fscanf(entree, "%s", &(options->fichier_opt));
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");

   for (i = 1; i <= 3; i++)
      {fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");}
   fscanf(entree, "%d", &(options->TRACE_STAT));
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");
   fscanf(entree, "%d", &(options->FICHIER_STAT));
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");
   fscanf(entree, "%s", &(options->fichier_stat));
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");

   for (i = 1; i <= 3; i++)
      {fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");}
   fscanf(entree, "%lg", &options->alpha);
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");
   fscanf(entree, "%lg", &options->beta);
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");
   fscanf(entree, "%lg", &options->gamma);
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");
   fscanf(entree, "%lg", &options->delta);
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");

   for (i = 1; i <= 3; i++)
      {fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");}
   fscanf(entree, "%d", &(options->type_initialisation));
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");
   fscanf(entree, "%lg", &options->param_initialisation);
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");

   for (i = 1; i <= 3; i++)
      {fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");}
   fscanf(entree, "%d", &(options->type_pivot));
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");

   for (i = 1; i <= 3; i++)
      {fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");}
   fscanf(entree, "%d", &(options->type_critere_arret));
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");
   fscanf(entree, "%lg", &options->precision_cout);
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");
   fscanf(entree, "%lg", &options->precision_X);
   fscanf(entree, "%*[^\n]"); fscanf(entree, "%*c");
   fscanf(entree, "%d", &(options->max_eval_cout));
   fclose(entree);
   }
   else
      {
      /*printf("\n!!! ATTENTION: le fichier nelder.dat n'est pas dans le repertoire courant !!!\n");*/
      Nelder_options_def(options);
      }

}


/*********************************************************************/
/*                                                                   */
/*                     --- Nelder_options_def ---                    */
/*                                                                   */
/* fct: initialise les options par defaut du simplexe.               */
/*                                                                   */
/* ver: 17 fevrier 1994                                              */
/*                                                                   */
/*********************************************************************/

void Nelder_options_def
(
Nelder_options *options   /* ptr du vecteur des options */
)
{
   options->TRACE = FALSE;
   options->TRACE_FICHIER = FALSE;
   options->fichier_trace[0] = 'n';
   options->fichier_trace[1] = 'e';
   options->fichier_trace[2] = 'l';
   options->fichier_trace[3] = 'd';
   options->fichier_trace[4] = 'e';
   options->fichier_trace[5] = 'r';
   options->fichier_trace[6] = '.';
   options->fichier_trace[7] = 'o';
   options->fichier_trace[8] = 'u';
   options->fichier_trace[9] = 't';
   options->fichier_trace[10] = '\0';

   options->TRACE_OPT = TRUE;
   options->FICHIER_OPT = FALSE;
   options->fichier_opt[0] = 'n';
   options->fichier_opt[1] = 'e';
   options->fichier_opt[2] = 'l';
   options->fichier_opt[3] = 'd';
   options->fichier_opt[4] = 'e';
   options->fichier_opt[5] = 'r';
   options->fichier_opt[6] = '.';
   options->fichier_opt[7] = 'o';
   options->fichier_opt[8] = 'p';
   options->fichier_opt[9] = 't';
   options->fichier_opt[10] = '\0';

   options->TRACE_STAT = TRUE;
   options->FICHIER_STAT = FALSE;
   options->fichier_stat[0] = 'n';
   options->fichier_stat[1] = 'e';
   options->fichier_stat[2] = 'l';
   options->fichier_stat[3] = 'd';
   options->fichier_stat[4] = 'e';
   options->fichier_stat[5] = 'r';
   options->fichier_stat[6] = '.';
   options->fichier_stat[7] = 's';
   options->fichier_stat[8] = 't';
   options->fichier_stat[9] = 'a';
   options->fichier_stat[10] = 't';
   options->fichier_stat[11] = '\0';

   options->alpha = 1.0;
   options->beta = 0.5;
   options->gamma = 2.0;
   options->delta = 2.0;
   options->type_initialisation = 1;
   options->param_initialisation = 0.10;
   options->type_pivot = 1;
   options->type_critere_arret = 1;
   options->precision_cout = 1.0e-6;
   options->precision_X = 1.0e-6;
   options->max_eval_cout = 1000;
}


/*********************************************************************/
/*                                                                   */
/*                    --- Nelder_optimisation ---                    */
/*                                                                   */
/* fct: optimisation non-lineaire avec le simplexe de Nelder-Mead.   */
/*                                                                   */
/* car: les parametres complementaires au calcul du cout sont passes */
/*      comme variables globales dans la fonction: Nelder_cout       */
/*                                                                   */
/* ver: 12 decembre 1993                                             */
/*                                                                   */
/*********************************************************************/

void Nelder_optimisation
(
Nelder_options *options,   /* ptr des options du simplexe           */
double (*Nelder_cout)      /* ptr de la fonction de calcul du cout  */
       (
       int    *n,      /* ptr du nombre de parametres de commande    */
       double *X,      /* ptr du vecteur des parametres de commande  */
       double *Xmin,   /* ptr des bornes minimales du domaine des X  */
       double *Xmax,   /* ptr des bornes maximales du domaine des X  */
       double *duser,  /* ptr de doubles (variables utilisateur)     */
       int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
       int    *iret,   /* ptr du code de retour des calculs (0 = ok) */
       int    *ier     /* ptr du code d'erreur des calculs (0 = ok)  */
       ),
int    *n,      /* ptr du nombre de parametres de commande   */
double *X,      /* ptr du vecteur des parametres de commande */
double *Xmin,   /* ptr des bornes minimales du domaine des X */
double *Xmax,   /* ptr des bornes maximales du domaine des X */
double *duser,  /* ptr de doubles (variables utilisateur)     */
int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
int    *ier     /* ptr du code d'erreur de l'optimisation:   */
		/*	      + = nb d'evaluations du cout   */
		/*	     -1 = max eval. cout atteint     */
		/*	    -99 = erreur dans Nelder_cout    */
)
{
   FILE *fich_trace = 0;
   int  i, j, iret_cout, ier_cout;
   int  etat_calcul, nb_iter_simplexe, nb_eval_cout;
   double *delta_X;
   Nelder_point    *pivot, *reflexion, *expansion, *contraction, *temp;
   Nelder_simplexe *simplexe;

/*** allocation dynamique de la memoire ***/

   delta_X        = (double *) malloc(*n * sizeof(double));			   

   pivot          = (Nelder_point *) malloc(sizeof(Nelder_point));
   pivot->n       = *n;
   pivot->x       = (double *) malloc(pivot->n * sizeof(double));
   reflexion      = (Nelder_point *) malloc(sizeof(Nelder_point));
   reflexion->n   = *n;
   reflexion->x   = (double *) malloc(reflexion->n * sizeof(double));
   expansion      = (Nelder_point *) malloc(sizeof(Nelder_point));
   expansion->n   = *n;
   expansion->x   = (double *) malloc(expansion->n * sizeof(double));
   contraction    = (Nelder_point *) malloc(sizeof(Nelder_point));
   contraction->n = *n;
   contraction->x = (double *) malloc(contraction->n * sizeof(double));

   simplexe    = (Nelder_simplexe *) malloc(sizeof(Nelder_simplexe));
   simplexe->dim = *n + 1;
   simplexe->pt = (Nelder_point **) malloc(simplexe->dim * sizeof(Nelder_point *));
   for (i = 0; i < simplexe->dim; i++)
      {
      simplexe->pt[i] = (Nelder_point *) malloc(simplexe->dim * sizeof(Nelder_point));
      simplexe->pt[i]->n = *n;
      simplexe->pt[i]->x = (double *) malloc(simplexe->pt[i]->n * sizeof(double));
      }

/*** initialisation du flux de sortie ***/

   if (options->TRACE)
      {
      if (options->TRACE_FICHIER)
         {
         fich_trace = fopen(options->fichier_trace, "w");
         if (fich_trace)
            {}
         else
            {
            printf("\n!!! ATTENTION: impossible d'ouvrir le fichier: %s !!!\n",&(options->fichier_trace));
            fich_trace = stdout;
            }
         }
      else
         fich_trace = stdout;
      }

/*** initialisation des variables ***/

   for (j = 0; j < *n; j++)
      delta_X[j] = Xmax[j] - Xmin[j];

/*** initialisation du simplexe ***/

   simplexe->min = 0;
   simplexe->max = simplexe->dim - 1;
   
   nb_eval_cout = 0;
   Nelder_initial(options->type_initialisation, options->param_initialisation,
                  X, Xmin, Xmax, simplexe);
   for (i = 0; i < simplexe->dim; i++)
      {
      simplexe->pt[i]->fx = Nelder_cout(n,simplexe->pt[i]->x,Xmin,Xmax,duser,iuser,&iret_cout,&ier_cout);
      if (ier_cout < 0) {*ier = ERREUR_CALCUL_COUT; return;}
      nb_eval_cout++;
      }

/*** deplacement du simplexe ***/

   if (options->TRACE) Nelder_trace_initial(fich_trace, simplexe);
   nb_iter_simplexe = 0;
   etat_calcul = EN_COURS;

   while (etat_calcul == EN_COURS)
      {
      nb_iter_simplexe++;
      Nelder_tri_simplexe(simplexe);
      if (options->TRACE) Nelder_trace(fich_trace, simplexe);
      Nelder_pivot(options->type_pivot, simplexe, pivot);

      Nelder_reflexion(options->alpha, pivot, simplexe->pt[simplexe->max], reflexion, Xmin, Xmax);
      reflexion->fx = Nelder_cout(n,reflexion->x,Xmin,Xmax,duser,iuser,&iret_cout,&ier_cout);
      if (ier_cout < 0) {*ier = ERREUR_CALCUL_COUT; return;}
      nb_eval_cout++;

      if (reflexion->fx < simplexe->pt[simplexe->min]->fx)
         {
         Nelder_expansion(options->gamma, pivot, reflexion, expansion, Xmin, Xmax);
         expansion->fx = Nelder_cout(n,expansion->x,Xmin,Xmax,duser, iuser,&iret_cout,&ier_cout);
         if (ier_cout < 0) {*ier = ERREUR_CALCUL_COUT; return;}
         nb_eval_cout++;

         if (expansion->fx < reflexion->fx)
            {
            temp = simplexe->pt[simplexe->max];
            simplexe->pt[simplexe->max] = expansion;
            expansion = temp;
            if (options->TRACE) fprintf(fich_trace, "---> expansion \n");
            }
         else
            {
            temp = simplexe->pt[simplexe->max];
            simplexe->pt[simplexe->max] = reflexion;
            reflexion = temp;
            if (options->TRACE) fprintf(fich_trace, "---> reflexion \n");
            }
         }
      else if (reflexion->fx < simplexe->pt[simplexe->max-1]->fx)
         {
         temp = simplexe->pt[simplexe->max];
         simplexe->pt[simplexe->max] = reflexion;
         reflexion = temp;
         if (options->TRACE) fprintf(fich_trace, "---> reflexion \n");
         }
      else
         {
         if (reflexion->fx < simplexe->pt[simplexe->max]->fx)
            {
            temp = simplexe->pt[simplexe->max];
            simplexe->pt[simplexe->max] = reflexion;
            reflexion = temp;
            if (options->TRACE) fprintf(fich_trace, "---> reflexion \n");
            }
         Nelder_contraction(options->beta, pivot, simplexe->pt[simplexe->max], contraction, Xmin, Xmax);
         contraction->fx = Nelder_cout(n,contraction->x,Xmin,Xmax,duser,iuser,&iret_cout,&ier_cout);
         if (ier_cout < 0) {*ier = ERREUR_CALCUL_COUT; return;}
         nb_eval_cout++;

         if (contraction->fx < simplexe->pt[simplexe->max]->fx)
            {
            temp = simplexe->pt[simplexe->max];
            simplexe->pt[simplexe->max] = contraction;
            contraction = temp;
            if (options->TRACE) fprintf(fich_trace, "---> contraction \n");
            }
         else
            {
            Nelder_reduction(options->delta, simplexe);
            for (i = 0; i < simplexe->dim; i++)
               if (i != simplexe->min)
                  {
                  simplexe->pt[i]->fx = Nelder_cout(n,simplexe->pt[i]->x,Xmin,Xmax,duser,iuser,&iret_cout,&ier_cout);
                  if (ier_cout < 0) {*ier = ERREUR_CALCUL_COUT; return;}
                  nb_eval_cout++;
                  }
            if (options->TRACE) fprintf(fich_trace, "---> reduction \n");
            }
         }

      if (Nelder_critere_arret(options->type_critere_arret, options->precision_cout,
                               options->precision_X, delta_X, simplexe))
         etat_calcul = MINIMUM_ATTEINT;
      else if (nb_eval_cout >= (options->max_eval_cout * *n))
         etat_calcul = ERREUR_MAX_EVAL_COUT;
      else
         etat_calcul = EN_COURS;
      }

   Nelder_tri_simplexe(simplexe);
   if (options->TRACE)
      {
      Nelder_trace(fich_trace, simplexe);
      Nelder_trace_final(fich_trace, etat_calcul, nb_iter_simplexe, nb_eval_cout);
      }
   for (j = 0; j < *n; j++)
      X[j] = simplexe->pt[simplexe->min]->x[j];

   if (etat_calcul == MINIMUM_ATTEINT)
      *ier = nb_eval_cout;
   else
      *ier = etat_calcul;

   if (options->TRACE && options->TRACE_FICHIER) fclose(fich_trace);

/*** liberation dynamique de la memoire ***/

   free(delta_X);
   free(pivot->x);
   free(pivot);
   free(reflexion->x);
   free(reflexion);
   free(expansion->x);
   free(expansion);
   free(contraction->x);
   free(contraction);

   for (i = 0; i < simplexe->dim; i++)
      {
      free(simplexe->pt[i]->x);
      free(simplexe->pt[i]);
      }
   free(simplexe->pt);
   free(simplexe);
}


/*********************************************************************/
/*                                                                   */
/*                         --- opt_nelder ---                        */
/*                                                                   */
/* fct: optimisation non-lineaire avec le simplexe de Nelder-Mead.   */
/*                                                                   */
/* car: les parametres complementaires au calcul du cout sont passes */
/*      comme variables globales dans la fonction: Nelder_cout       */
/*                                                                   */
/* ver: 4 fevrier 1994                                               */
/*                                                                   */
/*********************************************************************/

void opt_nelder
(
double (*Nelder_cout)   /* ptr de la fonction de calcul du cout */
       (
       int    *n,      /* ptr du nombre de parametres de commande    */
       double *X,      /* ptr du vecteur des parametres de commande  */
       double *Xmin,   /* ptr des bornes minimales du domaine des X  */
       double *Xmax,   /* ptr des bornes maximales du domaine des X  */
       double *duser,  /* ptr de doubles (variables utilisateur)     */
       int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
       int    *iret,   /* ptr du code de retour des calculs (0 = ok) */
       int    *ier     /* ptr du code d'erreur des calculs (0 = ok)  */
       ),
int    *n,      /* ptr du nombre de parametres de commande   */
double *X,      /* ptr du vecteur des parametres de commande */
double *Xmin,   /* ptr des bornes minimales du domaine des X */
double *Xmax,   /* ptr des bornes maximales du domaine des X */
double *duser,  /* ptr de doubles (variables utilisateur)     */
int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
int    *ier     /* ptr du code d'erreur de l'optimisation:   */
		/*	      + = nb d'evaluations du cout   */
		/*	     -1 = max eval. cout atteint     */
		/*	    -99 = erreur dans Nelder_cout    */
)
{
   Nelder_options options;

   Nelder_init_options(&options);
   Nelder_optimisation(&options, Nelder_cout, n, X, Xmin, Xmax, duser, iuser, ier);
}


/*********************************************************************/
/*                                                                   */
/*                         --- mopt_nelder ---                       */
/*                                                                   */
/* fct: optimisation non-lineaire avec le simplexe de Nelder-Mead    */
/*      a partir de plusieurs points de depart initiaux.             */
/*                                                                   */
/* car: - le premier point de depart de l'initialisation est defini  */
/*        par la vecteur X, les autres sont choisis de facon alea-   */
/*        toire a l'interieur du domaine d'admissibilite [Xmin,Xmax] */
/*                                                                   */
/*      - relance systematique de l'optimisation a chaque point      */
/*        optimal pour verifier sa validite.                         */
/*                                                                   */
/*      - les parametres complementaires au calcul du cout sont pas- */
/*        ses comme variables globales dans la fonction: Nelder_cout */
/*                                                                   */
/* ver: 17 fevrier 1994                                              */
/*                                                                   */
/*********************************************************************/

void mopt_nelder
(
int    *nb_opt,        /* ptr du nb de points de depart de l'optim.  */
double (*Nelder_cout)  /* ptr de la fonction de calcul du cout       */
       (
       int    *n,      /* ptr du nombre de parametres de commande    */
       double *X,      /* ptr du vecteur des parametres de commande  */
       double *Xmin,   /* ptr des bornes minimales du domaine des X  */
       double *Xmax,   /* ptr des bornes maximales du domaine des X  */
       double *duser,  /* ptr de doubles (variables utilisateur)     */
       int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
       int    *iret,   /* ptr du code de retour des calculs (0 = ok) */
       int    *ier     /* ptr du code d'erreur des calculs (0 = ok)  */
       ),
int    *n,      /* ptr du nombre de parametres de commande   */
double *X,      /* ptr du vecteur des parametres de commande */
double *Xmin,   /* ptr des bornes minimales du domaine des X */
double *Xmax,   /* ptr des bornes maximales du domaine des X */
double *duser,  /* ptr de doubles (variables utilisateur)     */
int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
int    *ier     /* ptr du code d'erreur de l'optimisation:   */
		/*	      0 = ok			     */
		/*	     -1 = aucun d'optimum trouve     */
		/*	    -99 = erreur dans Nelder_cout    */
)
{
   Nelder_options options;
   int iopt, ix, init_ran0;
   int iret_cout, ier_cout, ier_opt;
   double cout_ini, cout=0, cout_opt=0;
   double *Xini, *Xopt, *delta_X;

   FILE *fich_opt=0, *fich_stat=0;
   boolean OPT_TROUVE;
   double pourcentage_tot, pourcentage_glo, pourcentage_res;
   double pourcentage_vio, pourcentage_max, pourcentage_imp;
   int nb_tot, nb_glo=0, nb_res=0, nb_vio=0, nb_max=0, nb_imp=0;
   int nb_eval_tot, nb_eval_glo=0, nb_eval_res=0;
   int nb_eval_vio=0, nb_eval_max=0, nb_eval_imp=0;
   int moy_eval_tot, moy_eval_glo, moy_eval_res;
   int moy_eval_vio, moy_eval_max, moy_eval_imp;
   int nb_eval_cout1=0, nb_eval_cout2=0;

   boolean opt_connu;
   int     iopt_connu, total_opt_connu=0;
   int     *nb_opt_connu, *index_opt_connu;
   double  *cout_opt_connu, **Xopt_connu;


/*** allocation dynamique de la memoire ***/

   Xini = (double *) malloc(*n * sizeof(double));
   Xopt = (double *) malloc(*n * sizeof(double));
   delta_X = (double *) malloc(*n * sizeof(double));

   nb_opt_connu    = (int *) malloc(*nb_opt * sizeof(int));
   index_opt_connu = (int *) malloc(*nb_opt * sizeof(int));
   cout_opt_connu  = (double *) malloc(*nb_opt * sizeof(double));
   Xopt_connu      = (double **) malloc(*nb_opt * sizeof(double *));

/*** initialisation des variables ***/

   init_ran0 = 0;
   for (ix = 0; ix < *n; ix++)
      delta_X[ix] = Xmax[ix] - Xmin[ix];

/*** initialisation des options de l'algorithme d'optimisation ***/

   Nelder_init_options(&options);

/*** initialisation des flux de sortie ***/

   if (options.TRACE_OPT) {
      if (options.FICHIER_OPT)
         {
         fich_opt = fopen(options.fichier_opt, "w");
         if (fich_opt)
            {}
         else
            {
            printf("\n!!! ATTENTION: impossible d'ouvrir le fichier: %s !!!\n",&(options.fichier_opt));
            fich_opt = stdout;
            }
         }
      else
         fich_opt = stdout;
   }
   if (options.TRACE_STAT) {
      if (options.FICHIER_STAT)
         {
         fich_stat = fopen(options.fichier_stat, "w");
         if (fich_stat)
            {}
         else
            {
            printf("\n!!! ATTENTION: impossible d'ouvrir le fichier: %s !!!\n",&(options.fichier_stat));
            fich_stat = stdout;
            }
         }
      else
         fich_stat = stdout;
   }

/*** initialisation des statistiques ***/

   if (options.TRACE_STAT)
      {
      nb_tot = nb_glo = nb_res = 0;
      nb_vio = nb_max = nb_imp = 0;
      nb_eval_tot = nb_eval_glo = nb_eval_res = 0;
      nb_eval_vio = nb_eval_max = nb_eval_imp = 0;
      total_opt_connu = 0;
      }

/*** debut de l'optimisation ***/

   OPT_TROUVE = FALSE;
   for (iopt = 1; iopt <= *nb_opt; iopt++)
      {

   /*** initialisation du point de depart de l'optimisation ***/

      if (iopt == 1)
         COPY_DOUBLE(Xini, X, *n);
      else
         for (ix = 0; ix < *n; ix++)
            Xini[ix] = Xmin[ix] + delta_X[ix]*Optim_ran0(&init_ran0);

      cout_ini = Nelder_cout(n, Xini, Xmin, Xmax, duser, iuser, &iret_cout, &ier_cout);
      if (ier_cout < 0) {*ier = ERREUR_CALCUL_COUT; return;}

      if (options.TRACE_OPT)
         {
         fprintf(fich_opt,"\n---> Optimisation no. %1d: \n\n",iopt);
         for (ix = 0; ix < *n; ix++)
            fprintf(fich_opt,"X_ini(%1d) = %15.8g \n",ix+1,Xini[ix]);
         fprintf(fich_opt,"Cout_ini = %15.8g \n\n",cout_ini);
         }

   /*** 1ere optimisation a partir du point Xini ***/

      Nelder_optimisation(&options, Nelder_cout, n, Xini, Xmin, Xmax, duser, iuser, &ier_opt);
      if (ier_opt == ERREUR_MAX_EVAL_COUT)
         fprintf(fich_opt,"*** maximum d'evaluations de la fonction cout atteint ***\n\n");
      else if (ier_opt < 0)
         {*ier = ERREUR_CALCUL_COUT; return;}
      else
         {
         nb_eval_cout1 = ier_opt;

         cout = Nelder_cout(n, Xini, Xmin, Xmax, duser, iuser, &iret_cout, &ier_cout);
         if (ier_cout < 0) {*ier = ERREUR_CALCUL_COUT; return;}

         if (options.TRACE_OPT)
            {
            for (ix = 0; ix < *n; ix++)
               fprintf(fich_opt,"X_opt(%1d) = %15.8g \n",ix+1,Xini[ix]);
            fprintf(fich_opt,"Cout_opt = %15.8g \n\n",cout);
            }

      /*** relance de l'optimisation au point optimal ***/

         Nelder_optimisation(&options, Nelder_cout, n, Xini, Xmin, Xmax, duser, iuser, &ier_opt);
         if (ier_opt == ERREUR_MAX_EVAL_COUT)
	   if (options.TRACE_OPT)
	      fprintf(fich_opt,"*** maximum d'evaluations de la fonction cout atteint ***\n\n");
	   else
	     printf("*** maximum d'evaluations de la fonction cout atteint ***\n\n");
         else if (ier_opt < 0)
            {*ier = ERREUR_CALCUL_COUT; return;}
         else
            {
            nb_eval_cout2 = ier_opt;
            cout = Nelder_cout(n, Xini, Xmin, Xmax, duser, iuser, &iret_cout, &ier_cout);
            if (ier_cout < 0) {*ier = ERREUR_CALCUL_COUT; return;}

            if (options.TRACE_OPT)
               {
               fprintf(fich_opt,"Relance de l'optimisation: \n\n");
               for (ix = 0; ix < *n; ix++)
                  fprintf(fich_opt,"X_opt(%1d) = %15.8g \n",ix+1,Xini[ix]);
               fprintf(fich_opt,"Cout_opt = %15.8g \n\n",cout);
               }
         /*** verification de la commande optimale ***/

            if (OPT_TROUVE == FALSE)
               {
               OPT_TROUVE = TRUE;
               COPY_DOUBLE(Xopt, Xini, *n);
               cout_opt = cout;
               }
            else if (cout < cout_opt)
               {
               if (options.TRACE_STAT)
                  {
                  if (Optim_longvec_norm(*n,Xopt,Xini,delta_X)
                      > (options.precision_X*FACTEUR_PRECISION_OPTIMUM))
                     {
                     nb_res += nb_glo;
                     nb_glo = 0;
                     nb_eval_res += nb_eval_glo;
                     nb_eval_glo = 0;
                     }
                  }

               COPY_DOUBLE(Xopt, Xini, *n);
               cout_opt = cout;
               }
            }
         }

  /*** calcul des statistiques ***/

      if (options.TRACE_STAT)
         {
         if (ier_opt == ERREUR_MAX_EVAL_COUT)
            {
            nb_tot++;
            nb_max++;
            }
         else if (cout > INFINI)
            {
            nb_tot++;
            nb_imp++;
            }
         else 
            {
            if (iret_cout < 0)
               {
               nb_tot++;
               nb_vio++;
               nb_eval_vio += nb_eval_cout1 + nb_eval_cout2;
               }
            else if (Optim_longvec_norm(*n,Xopt,Xini,delta_X)
                     > (options.precision_X*FACTEUR_PRECISION_OPTIMUM))
               {
               nb_tot++;
               nb_res++;
               nb_eval_res += nb_eval_cout1 + nb_eval_cout2;
               }
            else
               {
               nb_tot++;
               nb_glo++;
               nb_eval_glo += nb_eval_cout1 + nb_eval_cout2;
               }

            opt_connu = FALSE;
            iopt_connu = 0;
            while ((total_opt_connu > 0) && (iopt_connu < total_opt_connu) && (! opt_connu))
               {
               if (Optim_longvec_norm(*n,Xopt_connu[iopt_connu],Xini,delta_X)
                     <= (options.precision_X*FACTEUR_PRECISION_OPTIMUM))
                  {
                  opt_connu = TRUE;
                  nb_opt_connu[iopt_connu]++;
                  }
               iopt_connu++;
               }
            if (! opt_connu)
               {
               Xopt_connu[total_opt_connu] = (double *) malloc(*n * sizeof(double));
               COPY_DOUBLE(Xopt_connu[total_opt_connu], Xini, *n);
               cout_opt_connu[total_opt_connu] = cout;
               nb_opt_connu[total_opt_connu] = 1;
               total_opt_connu++;
               }
            }
         }
      }

   if (OPT_TROUVE)
      {
      COPY_DOUBLE(X, Xopt, *n);

   /*** impression de la commande optimale ***/

      if (options.TRACE_OPT)
         {
         fprintf(fich_opt,"\n---> Commande optimale: \n\n");
         for (ix = 0; ix < *n; ix++)
            fprintf(fich_opt,"X_opt(%1d) = %15.8g \n",ix+1,Xopt[ix]);
         fprintf(fich_opt,"Cout_opt = %15.8g \n\n",cout_opt);
         }
      }
   else
      *ier = ERREUR_AUCUN_OPTIMUM;

/*** impression des statistiques ***/

   if (options.TRACE_STAT)
      {

   /*** affichage des statistiques de convergence ***/

      nb_tot = nb_glo + nb_res + nb_vio + nb_imp + nb_max;
      nb_eval_tot = nb_eval_glo + nb_eval_res + nb_eval_vio;
      if (nb_tot > 0)
         {
         pourcentage_glo = ((double) nb_glo/(double) nb_tot)*100.0;
         pourcentage_res = ((double) nb_res/(double) nb_tot)*100.0;
         pourcentage_vio = ((double) nb_vio/(double) nb_tot)*100.0;
         pourcentage_max = ((double) nb_max/(double) nb_tot)*100.0;
         pourcentage_imp = ((double) nb_imp/(double) nb_tot)*100.0;
         pourcentage_tot = ((double) nb_tot/(double) nb_tot)*100.0;
         }
      else
         {
         pourcentage_glo = 0.0;
         pourcentage_res = 0.0;
         pourcentage_vio = 0.0;
         pourcentage_max = 0.0;
         pourcentage_imp = 0.0;
         pourcentage_tot = 0.0;
         }
      if (nb_glo > 0)
         moy_eval_glo = (int)((double) nb_eval_glo/(double) nb_glo);
      else
         moy_eval_glo = 0;
      if (nb_res > 0)
         moy_eval_res = (int)((double) nb_eval_res/(double) nb_res);
      else
         moy_eval_res = 0;
      if (nb_vio > 0)
         moy_eval_vio = (int)((double) nb_eval_vio/(double) nb_vio);
      else
         moy_eval_vio = 0;
      if (nb_max > 0)
         moy_eval_max = (int)((double) nb_eval_max/(double) nb_max);
      else
         moy_eval_max = 0;
      if (nb_imp > 0)
         moy_eval_imp = (int)((double) nb_eval_imp/(double) nb_imp);
      else
         moy_eval_imp = 0;
      if ((nb_tot-nb_imp-nb_max) > 0)
         moy_eval_tot = (int)((double) nb_eval_tot/(double) (nb_tot-nb_imp-nb_max));
      else
         moy_eval_tot = 0;

      fprintf(fich_stat,"\n");
      fprintf(fich_stat,"+------------------------------------------+-------------------+-------------------+\n");
      fprintf(fich_stat,"|      STATISTIQUES DE L'OPTIMISATION      |    Nombre  (%%)    |  Moy. eval. cout  |\n");
      fprintf(fich_stat,"+------------------------------------------+-------------------+-------------------+\n");
      fprintf(fich_stat,"|  optimum global & contraintes respectees |  %5d  (%5.1f %%) |     %8d      |\n",
              nb_glo, pourcentage_glo, moy_eval_glo);
      fprintf(fich_stat,"|  optimum local & contraintes respectees  |  %5d  (%5.1f %%) |     %8d      |\n",
              nb_res, pourcentage_res, moy_eval_res);
      fprintf(fich_stat,"|  optimum local & contraintes violees     |  %5d  (%5.1f %%) |     %8d      |\n",
              nb_vio, pourcentage_vio, moy_eval_vio);
      fprintf(fich_stat,"|  maximum d'evaluations du cout atteint   |  %5d  (%5.1f %%) |      -------      |\n",
              nb_max, pourcentage_max);
      fprintf(fich_stat,"|  impossibilite physique                  |  %5d  (%5.1f %%) |      -------      |\n",
              nb_imp, pourcentage_imp);
      fprintf(fich_stat,"|                                          |                   |                   |\n");
      fprintf(fich_stat,"|  nombre total d'optimisations            |  %5d  (%5.1f %%) |     %8d      |\n",
              nb_tot, pourcentage_tot, moy_eval_tot);
      fprintf(fich_stat,"+------------------------------------------+-------------------+-------------------+\n");
      fprintf(fich_stat,"\n");

   /*** affichage des optimums ***/

      Optim_tri_index(total_opt_connu, cout_opt_connu, index_opt_connu);

      fprintf(fich_stat,"\n+------+  +-------------+  ");
      for (ix = 0; ix < *n; ix++) fprintf(fich_stat,"+-------------+  ");
      fprintf(fich_stat,"\n|  nb  |  |    cout     |  ");
      for (ix = 0; ix < *n; ix++) fprintf(fich_stat,"|      x%-4d  |  ",ix+1);
      fprintf(fich_stat,"\n+------+  +-------------+  ");
      for (ix = 0; ix < *n; ix++) fprintf(fich_stat,"+-------------+  ");
      fprintf(fich_stat,"\n");

      for (iopt_connu = 0; iopt_connu < total_opt_connu; iopt_connu++)
         {
         fprintf(fich_stat,"%7d  %15.8g  ", nb_opt_connu[index_opt_connu[iopt_connu]],
                                            cout_opt_connu[index_opt_connu[iopt_connu]]);
         for (ix = 0; ix < *n; ix++) fprintf(fich_stat,"%15.8g  ", 
                                             Xopt_connu[index_opt_connu[iopt_connu]][ix]);
         fprintf(fich_stat,"\n");
         free(Xopt_connu[index_opt_connu[iopt_connu]]);
         }
      fprintf(fich_stat,"\n");
      }

/*** fermeture des fichiers de sortie ***/

   if (options.TRACE_OPT && options.FICHIER_OPT)
      fclose(fich_opt);
   if (options.TRACE_STAT && options.FICHIER_STAT)
      fclose(fich_stat);

/*** liberation dynamique de la memoire ***/

   free(Xini);
   free(Xopt);
   free(delta_X);

   free(nb_opt_connu);
   free(index_opt_connu);
   free(cout_opt_connu);

   /****   for (i = 0; i < total_opt_connu; i++)
      {
      free(Xopt_connu[i]);
      }
      ***/
   free(Xopt_connu);
}


/*********************************************************************/
/*                                                                   */
/*                         --- opt_nelder_ ---                       */
/*                                                                   */
/* fct: optimisation non-lineaire avec le simplexe de Nelder-Mead.   */
/*                                                                   */
/* uti: exemple d'appel par un programme Fortran:                    */
/*         call opt_nelder(Nelder_cout, n, X, Xmin, Xmax, ier)       */
/*                                                                   */
/* car: les parametres complementaires au calcul du cout sont passes */
/*      comme variables globales dans la fonction: Nelder_cout       */
/*                                                                   */
/* ver: 12 decembre 1993                                             */
/*                                                                   */
/*********************************************************************/

void opt_nelder_
(
double (*Nelder_cout)   /* ptr de la fonction de calcul du cout */
       (
       int    *n,      /* ptr du nombre de parametres de commande    */
       double *X,      /* ptr du vecteur des parametres de commande  */
       double *Xmin,   /* ptr des bornes minimales du domaine des X  */
       double *Xmax,   /* ptr des bornes maximales du domaine des X  */
       double *duser,  /* ptr de doubles (variables utilisateur)     */
       int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
       int    *iret,   /* ptr du code de retour des calculs (0 = ok) */
       int    *ier     /* ptr du code d'erreur des calculs (0 = ok)  */
       ),
int    *n,      /* ptr du nombre de parametres de commande   */
double *X,      /* ptr du vecteur des parametres de commande */
double *Xmin,   /* ptr des bornes minimales du domaine des X */
double *Xmax,   /* ptr des bornes maximales du domaine des X */
double *duser,  /* ptr de doubles (variables utilisateur)     */
int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
int    *ier     /* ptr du code d'erreur de l'optimisation:   */
		/*	      + = nb d'evaluations du cout   */
		/*	     -1 = max eval. cout atteint     */
		/*	    -99 = erreur dans Nelder_cout    */
)
{
   opt_nelder(Nelder_cout, n, X, Xmin, Xmax, duser, iuser, ier);
}


/*********************************************************************/
/*                                                                   */
/*                         --- mopt_nelder_ ---                      */
/*                                                                   */
/* fct: optimisation non-lineaire avec le simplexe de Nelder-Mead    */
/*      a partir de plusieurs points de depart initiaux.             */
/*                                                                   */
/* uti: exemple d'appel par un programme Fortran:                    */
/*         call mopt_nelder(nb_opt,Nelder_cout,n,X,Xmin,Xmax,ier)    */
/*                                                                   */
/* car: - le premier point de depart de l'initialisation est defini  */
/*        par la vecteur X, les autres sont choisis de facon alea-   */
/*        toire a l'interieur du domaine d'admissibilite [Xmin,Xmax] */
/*                                                                   */
/*      - les parametres complementaires au calcul du cout sont pas- */
/*        ses comme variables globales dans la fonction: Nelder_cout */
/*                                                                   */
/* ver: 4 fevrier 1994                                               */
/*                                                                   */
/*********************************************************************/

void mopt_nelder_
(
int     *nb_opt,       /* ptr du nb de points de depart de l'optim.  */
double (*Nelder_cout)  /* ptr de la fonction de calcul du cout       */
       (
       int    *n,      /* ptr du nombre de parametres de commande    */
       double *X,      /* ptr du vecteur des parametres de commande  */
       double *Xmin,   /* ptr des bornes minimales du domaine des X  */
       double *Xmax,   /* ptr des bornes maximales du domaine des X  */
       double *duser,  /* ptr de doubles (variables utilisateur)     */
       int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
       int    *iret,   /* ptr du code de retour des calculs (0 = ok) */
       int    *ier     /* ptr du code d'erreur des calculs (0 = ok)  */
       ),
int    *n,      /* ptr du nombre de parametres de commande   */
double *X,      /* ptr du vecteur des parametres de commande */
double *Xmin,   /* ptr des bornes minimales du domaine des X */
double *Xmax,   /* ptr des bornes maximales du domaine des X */
double *duser,  /* ptr de doubles (variables utilisateur)     */
int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
int    *ier     /* ptr du code d'erreur de l'optimisation:   */
		/*	      0 = ok			     */
		/*	     -1 = aucun d'optimum trouve     */
		/*	    -99 = erreur dans Nelder_cout    */
)

{
   mopt_nelder(nb_opt, Nelder_cout, n, X, Xmin, Xmax, duser, iuser, ier);
}

/*** FIN ***/
