
/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef DATAMANAGER_H
#define DATAMANAGER_H

#include "../behav/Atom.h"
#include "../behav/State.h"
#include "../behav/StateVar.h"
#include "../behav/ObsVar.h"
#include "../behav/Comp.h"
#include "../behav/Trajectory.h"
#include "config.h"
#include "purge.h"
#include <string>
#include <vector>
#include <map>
#include <list>
#include <stack>
#include <fstream>
#include <sstream>

#if (__GNUC__ >= 3)
#include <ext/hash_map>
#else
#include <hash_map>
#endif

#if (__GNUC__ >= 3)
using __gnu_cxx::hash_map;
using __gnu_cxx::hash;
#else
using std::hash_map;
using std::hash;
#endif

class Formula;

#ifndef EQVECAT_S
#define EQVECAT_S
/* compare two unsorted vectors of atoms. */
struct eqvecat
{
  bool operator()(vector<Atom*>* v1, vector<Atom*>* v2) const
  {
    bool res = true;
    for (unsigned int i=0; i<v1->size(); i++) {
      for (unsigned int j=0; j<v2->size(); j++) {
	if ((*v1)[i] == (*v2)[j]) {
	  res = true;
	  j = v2->size();
	}
	else res = false;
      }
      if (!res) return false;
    }
    return true;
  }
};
#endif

#ifndef EQSTRING_S
#define EQSTRING_S
struct eqstring
{
  bool operator()(string s1, string s2)
  {
    return s1 == s2;
  }
};
#endif

/**
 * \class DataManager
 * \brief Manager for data of the model.
 */

class DataManager {

 public:
  bool running_status;
 
  static hash_map<const char*, Comp*, hash<const char*>, eqstr>* active_comps; /**< components in the model */
  static hash_map<const char*, Atom*, hash<const char*>, eqstr>* active_atoms; /**< atoms in the model */
  static hash_map<int, Transition*, hash<int> >* active_trans;  /**< transitions in the model */
  static vector<Clause*>* cls; /**< 'external clauses' in the model */
  //database of conflicts.
  //static hash_map<int, pair<vector<const char*>*, vector<const char*>* >*, hash<int> >* conflictsDB;
  /* overall trajectories cover = cover over all states in history. */
  static double traj_cover;
  /* cover on each state in history. */
  static hash_map<int, double, hash<int> >* states_cover;
  /* identity mapping function */
  static MapFunc* mpid;

  //testing efficiency.
  /*static int transitions_checked;
  static int trajectories_checked;
  static int trajectories_created;
  static int trajectories_destroyed;*/

  /* current model name */
  static string model_name;

#ifdef BENCHMK
  /* stream for benchmarks results storage. */
  static ofstream benchmk_stream;
  /* stream for itms benchmarks storage. */
  static ofstream itms_stream;
  /* counter for itms calls. */
  static unsigned int itms_calls;
#endif

  int total_clauses;

 public:
  DataManager(); //DataManager' structures are initialized while spawning the systemManager.
  ~DataManager();
  //DataManager instance();
  bool init();

  /************************** 
   *  Access to local data. * 
   **************************/
  void addComp(Comp* co); //add a component to the system
  Comp* getComp(const char* st);//retrieve the "active" component's address from the table
  void addAtom(Atom* at); //add an atom to the system
  Atom* getAtom(const char* id); //retrieve the "active" atom's address from the table
  /*void addStateVar(StateVar* sta);
    StateVar* getStateVar(const char* name);*/
  bool inVector(Atom* at, vector<Atom*>* vec); //test if at is in vec.
  bool inList(const char* elt, list<const char*>* li);
  //vector<Atom*>::iterator inVector2(Atom* at, vector<Atom*>* vec);
  bool inMultimap(Atom* at, multimap<string, Atom*, eqstring>* mult_map);

  /**
   * Probabilities management over state and trajectories.
   */
  void updateTrajectoriesCover();
  void updateStatesCover(int time_step, Transition* trans);

#ifdef BENCHMK
  /* benchmarking. */
  void addBmkEntry(int t_step, double time, bool diag, unsigned int traj_num);
  void addItmsBmkEntry(double time, unsigned int res_size, vector<const char*>* add_ats_names);
#endif
  
  /********************************************* 
   *   Viewers for the data and for debugging. * 
   *********************************************/
  void showComps();
  void showComp(const char* na);
  void graphComp (const char* na);
  void showAtoms();
  void printDebug(int count, ...);
  void viewMultimapAtomVector(multimap<double, vector<Atom*>*, greater<double> >* mapy); //debugging : view a multimap of vector of atoms... (operator is greater<double>).
  void viewActiveAtoms(bool bb); //debugging : view the active atoms.
  void viewAtomVector(vector<Atom*>* vec);
  void viewAtomVector2(vector<vector<Atom*>*>* vec);
  void viewMultimapAtom(multimap<string, Atom*, eqstring>* mlt_map); //show a multimap with atoms.
  void viewHashMapTrajEl(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* set);
  void viewHashMapTrajElLow(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* set);
  void viewTrueAtoms(multimap<string, Atom*, eqstring>* mlt_map);
  void viewTrueSol(multimap<double, multimap<string, Atom*, eqstring>*, greater<double> >* mlt_map);
  void viewTrueSol(multimap<double, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>*, greater<double> >* state);
  void viewTrajectoriesTransitionsMultimap(multimap<double, Trajectory*, greater<double> >* trajs);
  void viewTrajectoriesMultimap(multimap<double, Trajectory*, greater<double> >* trajs);
  void viewTrajectoriesMultimapFull(multimap<double, Trajectory*, greater<double> >* trajs);

  /* tools. */
  bool compareTrajsStates(State* st1, State* st2);
  void compareTrajs(multimap<double, Trajectory*, greater<double> >* res);
  void testTransitionsProbas();
  void dataError (long stade, long nerr);
};

#endif
